from typing import Any
from logging import Logger
from datetime import datetime
from pydantic import BaseModel, Field

from elysia.config import Settings
from elysia.config import settings as environment_settings
from elysia.objects import Result
from elysia.util.client import ClientManager
from elysia.util.parsing import format_dict_to_serialisable, remove_whitespace
from copy import deepcopy
from weaviate.classes.query import Filter


class Environment:
    """
    Store of all objects across different types of queries and responses.

    The environment is how the tree stores all objects across different types of queries and responses.
    This includes things like retrieved objects, retrieved metadata, retrieved summaries, etc.

    This is persistent across the tree, so that all agents are aware of the same objects.

    The environment is formatted and keyed as follows:
    ```python
    {
        "tool_name": {
            "result_name": [
                {
                    "metadata": dict,
                    "objects": list[dict],
                },
                ...
            ]
        }
    }
    ```

    Where `"tool_name"` is the name of the function that the result belongs to,
    e.g. if the result came from a tool called `"query"`, then `"tool_name"` is `"query"`.

    `"result_name"` is the name of the Result object, which can be customised,
    e.g. if the result comes from a specific collection, then `"result_name"` is the name of the collection.

    `"metadata"` is the metadata of the result,
    e.g. the time taken to retrieve the result, the query used, etc.

    `"objects"` is the list of objects retrieved from the result. This is a list of dictionaries, where each dictionary is an object.
    It is important that `objects` should have be list of dictionaries.
    e.g. each object that was returned from a retrieval, where the fields of each dictionary are the fields of the object returned.

    Each list under `result_name` is a dictionary with both `metadata` and `objects` keys.
    This is if, for example, you retrieve multiple objects from the same collection, each one is stored with different metadata.
    Because, for example, the query used to retrieve each object may be different (and stored differently in the metadata).

    The environment is initialised with a default "SelfInfo.generic" key, which is a list of one object, containing information about Elysia itself.

    You can use various methods to add, remove, replace, and find objects in the environment. See the methods below for more information.

    Within the environment, there is a variable called `hidden_environment`, which is a dictionary of key-value pairs.
    This is used to store information that is not shown to the LLM, but is instead a 'store' of data that can be used across tools.
    """

    def __init__(
        self,
        environment: dict[str, dict[str, Any]] | None = None,
        self_info: bool = True,
        hidden_environment: dict[str, Any] = {},
    ):
        if environment is None:
            environment = {}
        self.environment = environment
        self.hidden_environment = hidden_environment
        self.self_info = self_info
        if self_info:
            self.environment["SelfInfo"] = {}
            self.environment["SelfInfo"]["info"] = [
                {
                    "name": "Elysia",
                    "description": "An agentic RAG service in Weaviate.",
                    "purpose": remove_whitespace(
                        """Elysia is an agentic retrieval augmented generation (RAG) service, where users can query from Weaviate collections,
                    and the assistant will retrieve the most relevant information and answer the user's question. This includes a variety
                    of different ways to query, such as by filtering, sorting, querying multiple collections, and providing summaries
                    and textual responses.

                    Elysia will dynamically display retrieved objects from the collections in the frontend.
                    Elysia works via a tree-based approach, where the user's question is used to generate a tree of potential
                    queries to retrieve the most relevant information.
                    Each end of the tree connects to a separate agent that will perform a specific task, such as retrieval, aggregation, or generation, or more.

                    The tree itself has decision nodes that determine the next step in the query.
                    The decision nodes are decided via a decision-agent, which decides the task.

                    The agents communicate via a series of different prompts, which are stored in the prompt-library.
                    The decision-agent prompts are designed to be as general as possible, so that they can be used for a variety of different tasks.
                    Some of these variables include conversation history, retrieved objects, the user's original question, train of thought via model reasoning, and more.

                    The backend of Elysia is built with a range of libraries. Elysia is built with a lot of base python, and was constructed from the ground up to keep it as modular as possible.
                    The LLM components are built with DSPy, a library for training and running LLMs.
                    """
                    ),
                }
            ]

    def is_empty(self):
        """
        Check if the environment is empty.

        The "SelfInfo" key is not counted towards the empty environment.

        If the `.remove` method has been used, this is accounted for (e.g. empty lists count towards an empty environment).
        """
        empty = True
        for tool_key in self.environment.keys():
            if tool_key == "SelfInfo":
                continue

            for result_key in self.environment[tool_key].keys():
                if len(self.environment[tool_key][result_key]) > 0:
                    empty = False
                    break
        return empty

    def add(self, tool_name: str, result: Result, include_duplicates: bool = False):
        """
        Adds a result to the environment.
        Is called automatically by the tree when a result is returned from an agent.

        You can also add a result to the environment manually by using this method.
        In this case, you must be adding a 'Result' object (which has an implicit 'name' attribute used to key the environment).
        If you want to add something manually, you can use the `add_objects` method.

        Each item is added with a `_REF_ID` attribute, which is a unique identifier used to identify the object in the environment.
        If duplicate objects are detected, they are added with a duplicate `_REF_ID` entry detailing that they are a duplicate,
        as well as the `_REF_ID` of the original object.

        Args:
            tool_name (str): The name of the tool called that the result belongs to.
            result (Result): The result to add to the environment.
            include_duplicates (bool): Optional. Whether to include duplicate objects in the environment.
                Defaults to `False`, which still adds the duplicate object, but with a duplicate `_REF_ID` entry (and no repeating properties).
                If `True`, the duplicate object is added with a new `_REF_ID` entry, and the repeated properties are added to the object.
        """
        objects = result.to_json()
        name = result.name
        metadata = result.metadata
        if tool_name not in self.environment:
            self.environment[tool_name] = {}

        self.add_objects(tool_name, name, objects, metadata, include_duplicates)

    def add_objects(
        self,
        tool_name: str,
        name: str,
        objects: list[dict],
        metadata: dict = {},
        include_duplicates: bool = False,
    ):
        """
        Adds an object to the environment.
        Is not called automatically by the tree, so you must manually call this method.
        This is useful if you want to add an object to the environment manually that doesn't come from a Result object.

        Each item is added with a `_REF_ID` attribute, which is a unique identifier used to identify the object in the environment.
        If duplicate objects are detected, they are added with a duplicate `_REF_ID` entry detailing that they are a duplicate,
        as well as the `_REF_ID` of the original object.

        Args:
            tool_name (str): The name of the tool called that the result belongs to.
            name (str): The name of the result.
            objects (list[dict]): The objects to add to the environment.
            metadata (dict): Optional. The metadata of the objects to add to the environment.
                Defaults to an empty dictionary.
            include_duplicates (bool): Optional. Whether to include duplicate objects in the environment.
                Defaults to `False`, which still adds the duplicate object, but with a duplicate `_REF_ID` entry (and no repeating properties).
                If `True`, the duplicate object is added with a new `_REF_ID` entry, and the repeated properties are added to the object.

        """
        if tool_name not in self.environment:
            self.environment[tool_name] = {}

        if name not in self.environment[tool_name]:
            self.environment[tool_name][name] = []

        if len(objects) > 0:
            self.environment[tool_name][name].append(
                {
                    "metadata": metadata,
                    "objects": [],
                }
            )

            for i, obj in enumerate(objects):
                # check if the object is already in the environment
                obj_found = False
                where_obj = None
                for env_item in self.environment[tool_name][name]:
                    if obj in env_item["objects"]:
                        obj_found = True
                        where_obj = env_item["objects"].index(obj)
                        _REF_ID = env_item["objects"][where_obj]["_REF_ID"]
                        break

                if obj_found and not include_duplicates:
                    self.environment[tool_name][name][-1]["objects"].append(
                        {
                            "object_info": f"[repeat]",
                            "_REF_ID": _REF_ID,
                        }
                    )
                elif "_REF_ID" not in obj:
                    _REF_ID = f"{tool_name}_{name}_{len(self.environment[tool_name][name])}_{i}"
                    self.environment[tool_name][name][-1]["objects"].append(
                        {
                            "_REF_ID": _REF_ID,
                            **obj,
                        }
                    )
                else:
                    self.environment[tool_name][name][-1]["objects"].append(obj)

    def remove(self, tool_name: str, name: str, index: int | None = None):
        """
        Replaces the list of objects for the given `tool_name` and `name` with an empty list.

        Args:
            tool_name (str): The name of the tool called that the result belongs to.
            name (str): The name of the result.
            index (int | None): The index of the object to remove.
                If `None`, the entire list corresponding to `tool_name`/`name` is deleted.
                If an integer, the object at the given index is removed.
                Defaults to `None`.
                If `index=-1`, the last object is removed.
        """
        if tool_name in self.environment:
            if name in self.environment[tool_name]:
                if index is None:
                    self.environment[tool_name][name] = []
                else:
                    self.environment[tool_name][name].pop(index)

    def replace(
        self,
        tool_name: str,
        name: str,
        objects: list[dict],
        metadata: dict = {},
        index: int | None = None,
    ):
        """
        Replaces the list of objects for the given `tool_name` and `name` with the given list of objects.

        Args:
            tool_name (str): The name of the tool called that the result belongs to.
            name (str): The name of the result.
            objects (list[dict]): The objects to replace the existing objects with.
            metadata (dict): The metadata of the objects to replace the existing objects with.
            index (int | None): The index of the object to replace.
                If `None`, the entire list corresponding to `tool_name`/`name` is deleted and replaced with the new objects.
                If an integer, the object at the given index is replaced with the new objects.
                Defaults to `None`.
        """
        if tool_name in self.environment:
            if name in self.environment[tool_name]:
                if index is None:
                    self.environment[tool_name][name] = [
                        {
                            "metadata": metadata,
                            "objects": objects,
                        }
                    ]
                else:
                    self.environment[tool_name][name][index] = {
                        "metadata": metadata,
                        "objects": objects,
                    }

    def find(self, tool_name: str, name: str, index: int | None = None):
        """
        Finds a corresponding list of objects in the environment.
        Keyed via `tool_name` and `name`. See the base class description for more information on how the environment is keyed.

        Args:
            tool_name (str): The name of the tool called that the result belongs to.
            name (str): The name of the result.
            index (int | None): The index of the object to find.
                If `None`, the entire list corresponding to `tool_name`/`name` is returned.
                If an integer, the object at the given index is returned.

        Returns:
            (list[dict]): if `index` is `None` - The list of objects for the given `tool_name` and `name`.
            (dict): if `index` is an integer - The object at the given `index` for the given `tool_name` and `name`.
            (None): If the `tool_name` or `name` is not found in the environment.
        """

        if tool_name not in self.environment:
            return None
        if name not in self.environment[tool_name]:
            return None

        if index is None:
            return self.environment[tool_name][name]
        else:
            return self.environment[tool_name][name][index]

    def to_json(self, remove_unserialisable: bool = False):
        """
        Converts the environment to a JSON serialisable format.
        Used to access specific objects from the environment.
        """

        env_copy = deepcopy(self.environment)
        hidden_env_copy = deepcopy(self.hidden_environment)

        # Check if environment and hidden_environment are JSON serialisable
        for tool_name in env_copy:
            if tool_name != "SelfInfo":
                for name in self.environment[tool_name]:
                    for obj_metadata in self.environment[tool_name][name]:
                        format_dict_to_serialisable(
                            obj_metadata["metadata"], remove_unserialisable
                        )
                        for obj in obj_metadata["objects"]:
                            format_dict_to_serialisable(obj, remove_unserialisable)

        format_dict_to_serialisable(hidden_env_copy, remove_unserialisable)

        return {
            "environment": env_copy,
            "hidden_environment": hidden_env_copy,
            "self_info": self.self_info,
        }

    @classmethod
    def from_json(cls, json_data: dict):
        return cls(
            environment=json_data["environment"],
            hidden_environment=json_data["hidden_environment"],
            self_info=json_data["self_info"],
        )


def datetime_reference():
    date: datetime = datetime.now()
    return {
        "current_datetime": date.isoformat(),
        "current_day_of_week": date.strftime("%A"),
        "current_time_of_day": date.strftime("%I:%M %p"),
    }


class Atlas(BaseModel):
    style: str = Field(
        default="No style provided.",
        description="The writing style of the agent. "
        "This is the way the agent writes, and the tone of the language it uses.",
    )
    agent_description: str = Field(
        default="No description provided.",
        description="The description of the process you are following. This is pre-defined by the user. "
        "This could be anything - this is the theme of the program you are a part of.",
    )
    end_goal: str = Field(
        default="No end goal provided.",
        description="A short description of your overall goal. "
        "Use this to determine if you have completed your task. "
        "However, you can still choose to end actions early, "
        "if you believe the task is not possible to be completed with what you have available.",
    )
    datetime_reference: dict = Field(
        default=datetime_reference(),
        description="Current context information (e.g., date, time) for decision-making",
    )


class CollectionData:
    """
    Store of data about the Weaviate collections that are used in the tree.
    These are the output of the `preprocess` function.

    You do not normally need to interact with this class directly.
    Instead, do so via the `TreeData` class, which has corresponding methods to get the data in a variety of formats.
    (Such as via the `output_full_metadata` method.)
    """

    def __init__(
        self,
        collection_names: list[str],
        metadata: dict[str, Any] = {},
        logger: Logger | None = None,
    ):
        self.collection_names = collection_names
        self.metadata = metadata
        self.logger = logger

    async def set_collection_names(
        self, collection_names: list[str], client_manager: ClientManager
    ):
        temp_metadata = {}

        # check if any of these collections exist in the full metadata (cached)
        collections_to_get = []
        for collection_name in collection_names:
            if collection_name not in self.metadata:
                collections_to_get.append(collection_name)

        self.removed_collections = []
        self.incorrect_collections = []

        metadata_name = "ELYSIA_METADATA__"

        async with client_manager.connect_to_async_client() as client:
            # check if the metadata collection exists
            if not await client.collections.exists(metadata_name):
                self.removed_collections.extend(collections_to_get)
            else:
                metadata_collection = client.collections.get(metadata_name)
                filters = (
                    Filter.any_of(
                        [
                            Filter.by_property("name").equal(collection_name)
                            for collection_name in collections_to_get
                        ]
                    )
                    if len(collections_to_get) >= 1
                    else None
                )
                metadata = await metadata_collection.query.fetch_objects(
                    filters=filters,
                    limit=9999,
                )
                metadata_map = {
                    metadata_obj.properties["name"]: metadata_obj.properties
                    for metadata_obj in metadata.objects
                }

                for collection_name in collections_to_get:

                    if not await client.collections.exists(collection_name):
                        self.incorrect_collections.append(collection_name)
                        continue

                    if collection_name not in metadata_map:
                        self.removed_collections.append(collection_name)
                    else:
                        properties = metadata_map[collection_name]
                        format_dict_to_serialisable(properties)  # type: ignore
                        temp_metadata[collection_name] = properties

        if len(self.removed_collections) > 0 and self.logger:
            self.logger.warning(
                "The following collections have not been pre-processed for Elysia. "
                f"{self.removed_collections}. "
                "Ignoring these collections for now."
            )

        if len(self.incorrect_collections) > 0 and self.logger:
            self.logger.warning(
                "The following collections cannot be found in this Weaviate cluster. "
                f"{self.incorrect_collections}. "
                "These are being ignored for now. Please check that the collection names are correct."
            )

        self.collection_names = [
            collection_name
            for collection_name in collection_names
            if collection_name not in self.removed_collections
            and collection_name not in self.incorrect_collections
        ]

        # add to cached full metadata
        self.metadata = {
            **self.metadata,
            **{
                collection_name: temp_metadata[collection_name]
                for collection_name in temp_metadata
                if collection_name not in self.metadata
            },
        }

        return self.collection_names

    def output_full_metadata(
        self, collection_names: list[str] | None = None, with_mappings: bool = False
    ):
        if collection_names is None:
            collection_names = self.collection_names

        if with_mappings:
            return {
                collection_name: self.metadata[collection_name]
                for collection_name in collection_names
            }
        else:
            return {
                collection_name: {
                    k: v
                    for k, v in self.metadata[collection_name].items()
                    if k != "mappings"
                }
                for collection_name in collection_names
            }

    def output_collection_summaries(self, collection_names: list[str] | None = None):
        if collection_names is None:
            return {
                collection_name: self.metadata[collection_name]["summary"]
                for collection_name in self.collection_names
            }
        else:
            return {
                collection_name: self.metadata[collection_name]["summary"]
                for collection_name in collection_names
            }

    def output_mapping_lists(self):
        return {
            collection_name: list(self.metadata[collection_name]["mappings"].keys())
            for collection_name in self.collection_names
        }

    def output_mappings(self):
        return {
            collection_name: self.metadata[collection_name]["mappings"]
            for collection_name in self.collection_names
        }

    def to_json(self):
        return {
            "collection_names": self.collection_names,
            "metadata": self.metadata,
        }

    @classmethod
    def from_json(cls, json_data: dict, logger: Logger | None = None):
        return cls(
            collection_names=json_data["collection_names"],
            metadata=json_data["metadata"],
            logger=logger,
        )


class TreeData:
    """
    Store of data across the tree.
    This includes things like conversation history, actions, decisions, etc.
    These data are given to ALL agents, so every agent is aware of the stage of the decision processes.

    This also contains functions that process the data into an LLM friendly format,
    such as a string with extra description.
    E.g. the number of trees completed is converted into a string (i/N)
         and additional warnings if i is close to N.

    The TreeData has the following objects:

    - collection_data (CollectionData): The collection metadata/schema, which contains information about the collections used in the tree.
        This is the store of data that is saved by the `preprocess` function, and retrieved on initialisation of this object.
    - atlas (Atlas): The atlas, described in the Atlas class.
    - user_prompt (str): The user's prompt.
    - conversation_history (list[dict]): The conversation history stored in the current tree, of the form:
        ```python
        [
            {
                "role": "user" | "assistant",
                "content": str,
            },
            ...
        ]
        ```
    - environment (Environment): The environment, described in the Environment class.
    - tasks_completed (list[dict]): The tasks completed as a list of dictionaries.
        This is separate from the environment, as it separates what tasks were completed in each prompt in which order.
    - num_trees_completed (int): The current level of the decision tree, how many iterations have been completed so far.
    - recursion_limit (int): The maximum number of iterations allowed in the decision tree.
    - errors (dict): A dictionary of self-healing errors that have occurred in the tree. Keyed by the function name that caused the error.

    In general, you should not initialise this class directly.
    But you can access the data in this class to access the relevant data from the tree (in e.g. tool construction/usage).
    """

    def __init__(
        self,
        collection_data: CollectionData,
        atlas: Atlas,
        user_prompt: str | None = None,
        conversation_history: list[dict] | None = None,
        environment: Environment | None = None,
        tasks_completed: list[dict] | None = None,
        num_trees_completed: int | None = None,
        recursion_limit: int | None = None,
        settings: Settings | None = None,
    ):
        if settings is None:
            self.settings = environment_settings
        else:
            self.settings = settings

        # -- Base Data --
        if user_prompt is None:
            self.user_prompt = ""
        else:
            self.user_prompt = user_prompt

        if conversation_history is None:
            self.conversation_history = []
        else:
            self.conversation_history = conversation_history

        if environment is None:
            self.environment = Environment()
        else:
            self.environment = environment

        if tasks_completed is None:
            self.tasks_completed = []
        else:
            self.tasks_completed = tasks_completed

        if num_trees_completed is None:
            self.num_trees_completed = 0
        else:
            self.num_trees_completed = num_trees_completed

        if recursion_limit is None:
            self.recursion_limit = 3
        else:
            self.recursion_limit = recursion_limit

        # -- Atlas --
        self.atlas = atlas

        # -- Collection Data --
        self.collection_data = collection_data
        self.collection_names = []

        # -- Errors --
        self.errors: dict[str, list[str]] = {}
        self.current_task = None

    def set_property(self, property: str, value: Any):
        self.__dict__[property] = value

    def update_string(self, property: str, value: str):
        if property not in self.__dict__:
            self.__dict__[property] = ""
        self.__dict__[property] += value

    def update_list(self, property: str, value: Any):
        if property not in self.__dict__:
            self.__dict__[property] = []
        self.__dict__[property].append(value)

    def update_dict(self, property: str, key: str, value: Any):
        if property not in self.__dict__:
            self.__dict__[property] = {}
        self.__dict__[property][key] = value

    def delete_from_dict(self, property: str, key: str):
        if property in self.__dict__ and key in self.__dict__[property]:
            del self.__dict__[property][key]

    def soft_reset(self):
        self.previous_reasoning = {}

    def _update_task(self, task_dict, key, value):
        if value is not None:
            if key in task_dict:
                # If key already exists, append to it
                if isinstance(value, str):
                    task_dict[key] += "\n" + value
                elif isinstance(value, int) or isinstance(value, float):
                    task_dict[key] += value
                elif isinstance(value, list):
                    task_dict[key].extend(value)
                elif isinstance(value, dict):
                    task_dict[key].update(value)
                elif isinstance(value, bool):
                    task_dict[key] = value
            elif key not in task_dict:
                # If key does not exist, create it
                task_dict[key] = value

    def update_tasks_completed(
        self, prompt: str, task: str, num_trees_completed: int, **kwargs
    ):
        # search to see if the current prompt already has an entry for this task
        prompt_found = False
        task_found = False
        iteration_found = False

        for i, task_prompt in enumerate(self.tasks_completed):
            if task_prompt["prompt"] == prompt:
                prompt_found = True
                for j, task_j in enumerate(task_prompt["task"]):
                    if task_j["task"] == task:
                        task_found = True
                        task_i = j  # position of the task in the prompt

                    if task_j["iteration"] == num_trees_completed:
                        iteration_found = True

        # If the prompt is not found, add it to the list
        if not prompt_found:
            self.tasks_completed.append({"prompt": prompt, "task": [{}]})
            self.tasks_completed[-1]["task"][0]["task"] = task
            self.tasks_completed[-1]["task"][0]["iteration"] = num_trees_completed
            for kwarg in kwargs:
                self._update_task(
                    self.tasks_completed[-1]["task"][0], kwarg, kwargs[kwarg]
                )
            return

        # If the prompt is found but the task is not, add it to the list
        if prompt_found and not task_found:
            self.tasks_completed[-1]["task"].append({})
            self.tasks_completed[-1]["task"][-1]["task"] = task
            self.tasks_completed[-1]["task"][-1]["iteration"] = num_trees_completed
            for kwarg in kwargs:
                self._update_task(
                    self.tasks_completed[-1]["task"][-1], kwarg, kwargs[kwarg]
                )
            return

        # task already exists in this query, but the iteration is new
        if prompt_found and task_found and not iteration_found:
            self.tasks_completed[-1]["task"].append({})
            self.tasks_completed[-1]["task"][-1]["task"] = task
            self.tasks_completed[-1]["task"][-1]["iteration"] = num_trees_completed
            for kwarg in kwargs:
                self._update_task(
                    self.tasks_completed[-1]["task"][-1], kwarg, kwargs[kwarg]
                )
            return

        # If the prompt is found and the task is found, update the task
        if prompt_found and task_found:
            for kwarg in kwargs:
                self._update_task(
                    self.tasks_completed[-1]["task"][task_i], kwarg, kwargs[kwarg]
                )

    def set_current_task(self, task: str):
        self.current_task = task

    def get_errors(self):
        if self.current_task == "elysia_decision_node":
            return self.errors
        elif self.current_task is None or self.current_task not in self.errors:
            return []
        else:
            return self.errors[self.current_task]

    def clear_error(self, task: str):
        if task in self.errors:
            self.errors[task] = []

    def tasks_completed_string(self):
        """
        Output a nicely formatted string of the tasks completed so far, designed to be used in the LLM prompt.
        This is where the outputs of the `llm_message` fields are displayed.
        You can use this if you are interfacing with LLMs in tools, to help it understand the context of the tasks completed so far.

        Returns:
            (str): A separated and formatted string of the tasks completed so far in an LLM-parseable format.
        """
        out = ""
        for j, task_prompt in enumerate(self.tasks_completed):
            out += f"<prompt_{j+1}>\n"
            out += f"Prompt: {task_prompt['prompt']}\n"

            for i, task in enumerate(task_prompt["task"]):
                out += f"<task_{i+1}>\n"

                if "action" in task and task["action"]:
                    out += f"Chosen action: {task['task']} (this does not mean it has been completed, only that it was chosen) "
                    out += "(Use the environment to judge if a task is completed)"
                else:
                    out += f"Chosen subcategory: {task['task']} (this action has not been completed, this is only a subcategory)"

                if "error" in task and task["error"]:
                    out += f" (There was an error during this tool call)\n"
                else:
                    out += f" (Successfully completed)\n"

                for key in task:
                    if key != "task" and key != "action":
                        out += f"{key.capitalize()}: {task[key]}\n"

                out += f"</task_{i+1}>\n"
            out += f"</prompt_{j+1}>\n"

        return out

    async def set_collection_names(
        self, collection_names: list[str], client_manager: ClientManager
    ):
        self.collection_names = await self.collection_data.set_collection_names(
            collection_names, client_manager
        )
        return self.collection_names

    def tree_count_string(self):
        out = f"{self.num_trees_completed+1}/{self.recursion_limit}"
        if self.num_trees_completed == self.recursion_limit - 1:
            out += " (this is the last decision you can make before being cut off)"
        if self.num_trees_completed >= self.recursion_limit:
            out += " (recursion limit reached, write your full chat response accordingly - the decision process has been cut short, and it is likely the user's question has not been fully answered and you either haven't been able to do it or it was impossible)"
        return out

    def output_collection_metadata(
        self, collection_names: list[str] | None = None, with_mappings: bool = False
    ):
        """
        Outputs the full metadata for the given collection names.

        Args:
            with_mappings (bool): Whether to output the mappings for the collections as well as the other metadata.

        Returns:
            dict (dict[str, dict]): A dictionary of collection names to their metadata.
                The metadata are of the form:
                ```python
                {
                    # summary statistics of each field in the collection
                    "fields": list = [
                        field_name_1: dict = {
                            "description": str,
                            "range": list[float],
                            "type": str,
                            "groups": dict[str, str],
                            "mean": float
                        },
                        field_name_2: dict = {
                            ... # same fields as above
                        },
                        ...
                    ],

                    # mapping_1, mapping_2 etc refer to frontend-specific types that the AI has deemed appropriate for this data
                    # then the dict is to map the frontend fields to the data fields
                    "mappings": dict = {
                        mapping_1: dict = {
                            "frontend_field_1": "data_field_1",
                            "frontend_field_2": "data_field_2",
                            ...
                        },
                        mapping_2: dict = {
                            ... # same fields as above
                        },
                        ...,
                    },

                    # number of items in collection (float but just for consistency)
                    "length": float,

                    # AI generated summary of the dataset
                    "summary": str,

                    # name of collection
                    "name": str,

                    # what named vectors are available and their properties (if any)
                    "named_vectors": list = [
                        {
                            "name": str,
                            "enabled": bool,
                            "source_properties": list,
                            "description": str # defaults to empty
                        },
                        ...
                    ],

                    # some config settings relevant for queries
                    "index_properties": {
                        "isNullIndexed": bool,
                        "isLengthIndexed": bool,
                        "isTimestampIndexed": bool,
                    },
                }
                ```
                If `with_mappings` is `False`, then the mappings are not included.
                Each key in the outer level dictionary is a collection name.

        """

        if collection_names is None:
            collection_names = self.collection_names

        return self.collection_data.output_full_metadata(
            collection_names, with_mappings
        )

    def output_collection_return_types(self) -> dict[str, list[str]]:
        """
        Outputs the return types for the collections in the tree data.
        Essentially, this is a list of the keys that can be used to map the objects to the frontend.

        Returns:
            (dict): A dictionary of collection names to their return types.
                ```python
                {
                    collection_name_1: list[str],
                    collection_name_2: list[str],
                    ...,
                }
                ```
                Each of these lists is a list of the keys that can be used to map the objects to the frontend.
        """
        collection_return_types = self.collection_data.output_mapping_lists()
        out = {
            collection_name: collection_return_types[collection_name]
            for collection_name in self.collection_names
        }
        return out

    def to_json(self, remove_unserialisable: bool = False):
        out = {
            k: v
            for k, v in self.__dict__.items()
            if k not in ["collection_data", "atlas", "environment", "settings"]
        }
        out["collection_data"] = self.collection_data.to_json()
        out["atlas"] = self.atlas.model_dump()
        out["environment"] = self.environment.to_json(remove_unserialisable)
        out["settings"] = self.settings.to_json()
        return out

    @classmethod
    def from_json(cls, json_data: dict):
        settings = Settings.from_json(json_data["settings"])
        logger = settings.logger
        collection_data = CollectionData.from_json(json_data["collection_data"], logger)
        atlas = Atlas.model_validate(json_data["atlas"])
        environment = Environment.from_json(json_data["environment"])

        tree_data = cls(
            collection_data=collection_data,
            atlas=atlas,
            environment=environment,
            settings=settings,
        )
        for item in json_data:
            if item not in ["collection_data", "atlas", "environment", "settings"]:
                tree_data.set_property(item, json_data[item])
        return tree_data



from datetime import datetime
import json

from weaviate.collections.classes.aggregate import (
    AggregateGroupByReturn,
)
from weaviate.classes.aggregate import GroupByAggregate
from weaviate.classes.query import Filter, Metrics
from weaviate.util import generate_uuid5

from elysia.tree.tree import Tree
from elysia.util.parsing import format_datetime
from elysia.api.core.log import logger
import weaviate.classes.config as wc


async def create_feedback_collection(client):
    await client.collections.create(
        "ELYSIA_FEEDBACK__",
        properties=[
            # session data
            wc.Property(
                name="user_id",
                data_type=wc.DataType.TEXT,
            ),
            wc.Property(
                name="conversation_id",
                data_type=wc.DataType.TEXT,
            ),
            wc.Property(
                name="query_id",
                data_type=wc.DataType.TEXT,
            ),
            # feedback value (between -2, -1, 1, 2)
            wc.Property(
                name="feedback",
                data_type=wc.DataType.NUMBER,
            ),
            # track which models were used
            wc.Property(
                name="modules_used",
                data_type=wc.DataType.TEXT_ARRAY,
            ),
            # Tree data (except available_information)
            wc.Property(
                name="user_prompt",
                data_type=wc.DataType.TEXT,
            ),
            wc.Property(
                name="conversation_history",
                data_type=wc.DataType.OBJECT_ARRAY,
                nested_properties=[
                    wc.Property(
                        name="role",
                        data_type=wc.DataType.TEXT,
                    ),
                    wc.Property(
                        name="content",
                        data_type=wc.DataType.TEXT,
                    ),
                ],
            ),
            wc.Property(
                name="tasks_completed",
                data_type=wc.DataType.OBJECT_ARRAY,
                nested_properties=[
                    wc.Property(
                        name="prompt",
                        data_type=wc.DataType.TEXT,
                    ),
                    wc.Property(
                        name="tasks",
                        data_type=wc.DataType.OBJECT_ARRAY,
                        nested_properties=[
                            wc.Property(
                                name="task",
                                data_type=wc.DataType.TEXT,
                            ),
                            wc.Property(
                                name="reasoning",
                                data_type=wc.DataType.TEXT,
                            ),
                            wc.Property(
                                name="todo",
                                data_type=wc.DataType.TEXT,
                            ),
                            wc.Property(name="action", data_type=wc.DataType.BOOL),
                            wc.Property(
                                name="count",
                                data_type=wc.DataType.NUMBER,
                            ),
                            wc.Property(
                                name="extra_string",
                                data_type=wc.DataType.TEXT,
                            ),
                        ],
                    ),
                ],
            ),
            # Extra specifics
            wc.Property(
                name="route",
                data_type=wc.DataType.TEXT_ARRAY,
            ),
            wc.Property(
                name="action_information",
                data_type=wc.DataType.OBJECT_ARRAY,
                nested_properties=[
                    wc.Property(
                        name="collection_name",
                        data_type=wc.DataType.TEXT,
                    ),
                    wc.Property(
                        name="action_name",
                        data_type=wc.DataType.TEXT,
                    ),
                    wc.Property(
                        name="code",
                        data_type=wc.DataType.OBJECT,
                        nested_properties=[
                            wc.Property(
                                name="title",
                                data_type=wc.DataType.TEXT,
                            ),
                            wc.Property(
                                name="language",
                                data_type=wc.DataType.TEXT,
                            ),
                            wc.Property(
                                name="text",
                                data_type=wc.DataType.TEXT,
                            ),
                        ],
                    ),
                    wc.Property(
                        name="return_type",
                        data_type=wc.DataType.TEXT,
                    ),
                    wc.Property(
                        name="output_type",
                        data_type=wc.DataType.TEXT,
                    ),
                ],
            ),
            # metadata
            wc.Property(
                name="time_taken_seconds",
                data_type=wc.DataType.NUMBER,
            ),
            wc.Property(
                name="base_lm_used",
                data_type=wc.DataType.TEXT,
            ),
            wc.Property(
                name="complex_lm_used",
                data_type=wc.DataType.TEXT,
            ),
            wc.Property(
                name="feedback_date",
                data_type=wc.DataType.DATE,
            ),
            wc.Property(
                name="decision_time",
                data_type=wc.DataType.NUMBER,
            ),
            # dump training_updates as string
            wc.Property(name="training_updates", data_type=wc.DataType.TEXT),
        ],
        vectorizer_config=[
            wc.Configure.NamedVectors.text2vec_openai(
                name="user_prompt",
                model="text-embedding-3-small",
                source_properties=["user_prompt"],
                vector_index_config=wc.Configure.VectorIndex.hnsw(
                    quantizer=wc.Configure.VectorIndex.Quantizer.sq(),
                ),
            ),
        ],
    )
    logger.info("Feedback collection (ELYSIA_FEEDBACK__) created!")


async def create_feedback(
    user_id: str, conversation_id: str, query_id: str, feedback: int, tree: Tree, client
):
    if not await client.collections.exists("ELYSIA_FEEDBACK__"):
        await create_feedback_collection(client)

    feedback_collection = client.collections.get("ELYSIA_FEEDBACK__")

    history = tree.history[query_id]

    # new_tasks_completed = []
    # for task_prompt in history["tree_data"].tasks_completed:
    #     new_tasks_completed.append({})
    #     new_tasks_completed[-1]["prompt"] = task_prompt["prompt"]
    #     new_tasks_completed[-1]["task"] = [
    #         {
    #             "task": h["task"],
    #             "reasoning": h["reasoning"],
    #             "todo": h["todo"],
    #             "count": float(h["count"])
    #         }
    #     for h in task_prompt["tasks"]
    #     ]

    # ensure "action" is a bool in tasks_completed
    for task_prompt in history["tree_data"].tasks_completed:
        for task in task_prompt["task"]:
            task["action"] = bool(task["action"])

    date_now = datetime.now()
    # date_now = date_now - timedelta(days=1)

    properties = {
        "user_id": user_id,
        "conversation_id": conversation_id,
        "query_id": query_id,
        "feedback": int(feedback),
        "modules_used": list(
            set([h["module_name"] for h in history["training_updates"]])
        ),
        "user_prompt": history["tree_data"].user_prompt,
        "conversation_history": history["tree_data"].conversation_history,
        "tasks_completed": history["tree_data"].tasks_completed,
        "route": history["decision_history"],
        "action_information": history["action_information"],
        "time_taken_seconds": history["time_taken_seconds"],
        "decision_time": tree.tracker.get_average_time("decision_node"),
        "base_lm_used": tree.base_lm.model,
        "complex_lm_used": tree.complex_lm.model,
        "feedback_datetime": format_datetime(date_now),
        "feedback_date": format_datetime(
            date_now.replace(hour=0, minute=0, second=0, microsecond=0)
        ),
        "training_updates": json.dumps(history["training_updates"]),
        "initialisation": history["initialisation"],
    }

    # uuid is generated based on the user_id, conversation_id, query_id ONLY
    # so if the user re-selects the same feedback, it will be updated instead of added
    session_uuid = generate_uuid5(
        {
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
        }
    )

    try:
        if await feedback_collection.data.exists(session_uuid):
            await feedback_collection.data.update(
                properties=properties, uuid=session_uuid
            )
        else:
            await feedback_collection.data.insert(
                properties=properties, uuid=session_uuid
            )
    except Exception as e:
        logger.exception(f"Whilst inserting feedback to the feedback collection")


async def view_feedback(user_id: str, conversation_id: str, query_id: str, client):
    if not await client.collections.exists("ELYSIA_FEEDBACK__"):
        raise Exception("No feedback collection found")

    feedback_collection = client.collections.get("ELYSIA_FEEDBACK__")
    session_uuid = generate_uuid5(
        {
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
        }
    )

    feedback_object = await feedback_collection.query.fetch_object_by_id(
        uuid=session_uuid
    )

    return feedback_object.properties


async def remove_feedback(user_id: str, conversation_id: str, query_id: str, client):
    if not await client.collections.exists("ELYSIA_FEEDBACK__"):
        raise Exception("No feedback collection found")

    feedback_collection = client.collections.get("ELYSIA_FEEDBACK__")
    session_uuid = generate_uuid5(
        {
            "user_id": user_id,
            "conversation_id": conversation_id,
            "query_id": query_id,
        }
    )
    await feedback_collection.data.delete_by_id(uuid=session_uuid)


async def feedback_metadata(client, user_id: str):

    if not await client.collections.exists("ELYSIA_FEEDBACK__"):
        return {
            "error": "",
            "total_feedback": 0,
            "feedback_by_value": {
                "negative": 0,
                "positive": 0,
                "superpositive": 0,
            },
            "feedback_by_date": {},
        }

    feedback_collection = client.collections.get("ELYSIA_FEEDBACK__")

    all_aggregate = await feedback_collection.aggregate.over_all(total_count=True)
    total_feedback = all_aggregate.total_count

    feedback_values = {
        "negative": 0.0,
        "positive": 1.0,
        "superpositive": 2.0,
    }

    feedback_by_date = {}
    feedback_by_value = {}
    for feedback_name, feedback_value in feedback_values.items():

        filters = Filter.by_property("feedback").equal(feedback_value)

        # by value
        agg_feedback_count_i = await feedback_collection.aggregate.over_all(
            filters=filters, return_metrics=[Metrics("feedback").integer(count=True)]
        )
        feedback_by_value[feedback_name] = agg_feedback_count_i.properties[
            "feedback"
        ].count

        # by date
        agg_feedback_i = await feedback_collection.aggregate.over_all(
            group_by=GroupByAggregate(prop="feedback_date"),
            filters=filters,
            return_metrics=[Metrics("feedback").number(count=True)],
        )

        # if there is a property
        if isinstance(agg_feedback_i, AggregateGroupByReturn):

            for date_group in agg_feedback_i.groups:

                date_val = datetime.fromisoformat(date_group.grouped_by.value).strftime(  # type: ignore
                    "%Y-%m-%d"
                )

                if date_val not in feedback_by_date:
                    feedback_by_date[date_val] = {
                        "negative": 0,
                        "positive": 0,
                        "superpositive": 0,
                    }

                feedback_by_date[date_val][feedback_name] += date_group.properties[
                    "feedback"
                ].count  # type: ignore

    # fill in empties with zeros
    for date in feedback_by_date:
        for feedback_name in feedback_values:
            if feedback_name not in feedback_by_date[date]:
                feedback_by_date[date][feedback_name] = 0

    agg_feedback = await feedback_collection.aggregate.over_all(
        group_by=GroupByAggregate(prop="feedback_date"),
        return_metrics=[Metrics("feedback").number(mean=True, count=True)],
    )

    if isinstance(agg_feedback, AggregateGroupByReturn):

        for date_group in agg_feedback.groups:

            date_val = datetime.fromisoformat(date_group.grouped_by.value).strftime(  # type: ignore
                "%Y-%m-%d"
            )

            feedback_by_date[date_val]["mean"] = date_group.properties["feedback"].mean  # type: ignore
            feedback_by_date[date_val]["count"] = date_group.properties[
                "feedback"
            ].count  # type: ignore

    return {
        "total_feedback": total_feedback,
        "feedback_by_value": feedback_by_value,
        "feedback_by_date": feedback_by_date,
    }


# if __name__ == "__main__":
#     import os
#     import weaviate
#     from weaviate.classes.init import Auth
#     client = weaviate.connect_to_weaviate_cloud(
#         cluster_url = os.environ.get("WCD_URL"),
#         auth_credentials = Auth.api_key(os.environ.get("WCD_API_KEY")),
#         headers={"X-OpenAI-API-Key": os.environ.get("OPENAI_API_KEY")}
#     )
#     print(feedback_metadata(client))



from .collection import (
    preprocess,
    preprocess_async,
    view_preprocessed_collection,
    view_preprocessed_collection_async,
    edit_preprocessed_collection,
    edit_preprocessed_collection_async,
    delete_preprocessed_collection,
    delete_preprocessed_collection_async,
    preprocessed_collection_exists,
    preprocessed_collection_exists_async,
)



"""
Test more complex prompts that require multiple logical steps.
E.g. aggregation _and_ query.
Focusing only on whether the correct tools are called, not the accuracy of the results.
"""

import pytest
from copy import deepcopy
from deepeval import evaluate, metrics
from deepeval.test_case import LLMTestCase, LLMTestCaseParams, ToolCall

from elysia import Tree, Settings
from elysia.util.client import ClientManager
from elysia import configure, preprocess
from elysia.preprocessing.collection import preprocessed_collection_exists

configure(logging_level="DEBUG")


def get_tools_called(tree: Tree, user_prompt: str):
    tools_called = []
    if user_prompt in tree.actions_called:
        for action in tree.actions_called[user_prompt]:
            if action["name"] in tree.tools:
                tools_called.append(
                    ToolCall(
                        name=action["name"],
                        description=tree.tools[action["name"]].description,
                        input_parameters=action["inputs"],
                        output=action["output"] if "output" in action else [],
                    )
                )
    return tools_called


async def run_tree(user_prompt: str, collection_names: list[str]) -> Tree:

    tree = Tree()

    async for result in tree.async_run(
        user_prompt,
        collection_names=collection_names,
    ):
        pass
    return tree


@pytest.mark.asyncio
async def test_aggregation_and_query():
    """
    Test that the correct tools are called for a prompt that requires both aggregation and query.
    """

    # if the collection is not found, skip the test
    client_manager = ClientManager()
    if not client_manager.client.collections.exists("Weather"):
        pytest.skip("Collection not found in Weaviate")

    user_prompt = "What was the maximum weather in 2015? Then, what was the weather like in the surrounding week of that maximum?"
    tree = await run_tree(user_prompt, ["Weather"])

    all_decision_history = []
    for iteration in tree.decision_history:
        all_decision_history.extend(iteration)

    assert "aggregate" in all_decision_history
    assert "query" in all_decision_history

    # -- DeepEval tests --
    deepeval_metric = metrics.TaskCompletionMetric(
        threshold=0.5, model="gpt-4o-mini", include_reason=True
    )

    tools_called = get_tools_called(tree, user_prompt)

    test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        tools_called=tools_called,
    )

    res = evaluate(test_cases=[test_case], metrics=[deepeval_metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason


@pytest.mark.asyncio
async def test_query_with_chunking():
    """
    Test that the correct tools are called for a prompt that requires both aggregation and query.
    """

    try:
        # if the collection is not found, skip the test
        client_manager = ClientManager()
        if not client_manager.client.collections.exists(
            "Weaviate_documentation"
        ) or not client_manager.client.collections.exists("Weaviate_blogs"):
            pytest.skip("Collections not found in Weaviate")

        if not preprocessed_collection_exists("Weaviate_documentation", client_manager):
            preprocess(
                "Weaviate_documentation",
                client_manager,
            )
        if not preprocessed_collection_exists("Weaviate_blogs", client_manager):
            preprocess(
                "Weaviate_blogs",
                client_manager,
            )

        user_prompt = "How does Weaviate use HNSW?"
        tree = await run_tree(user_prompt, ["Weaviate_documentation", "Weaviate_blogs"])

        all_decision_history = []
        for iteration in tree.decision_history:
            all_decision_history.extend(iteration)

        assert "query" in all_decision_history
        env_key = (
            "summarise_items"
            if "summarise_items" in tree.tree_data.environment.environment
            else "query"
        )
        assert env_key in tree.tree_data.environment.environment

        collections_queried = list(
            tree.tree_data.environment.environment[env_key].keys()
        )

        for collection in collections_queried:
            assert collection in tree.tree_data.environment.environment[env_key]
            assert len(tree.tree_data.environment.environment[env_key][collection]) > 0
            for item in tree.tree_data.environment.environment[env_key][collection][0][
                "objects"
            ]:
                assert "chunk_spans" in item and item["chunk_spans"] is not []

        # -- DeepEval tests --
        deepeval_metric = metrics.TaskCompletionMetric(
            threshold=0.5, model="gpt-4o-mini", include_reason=True
        )

        tools_called = get_tools_called(tree, user_prompt)

        test_case = LLMTestCase(
            input=user_prompt,
            actual_output=tree.tree_data.conversation_history[-1]["content"],
            tools_called=tools_called,
        )

        res = evaluate(test_cases=[test_case], metrics=[deepeval_metric])
        for test_case in res.test_results:
            assert test_case.success, test_case.metrics_data[0].reason
    finally:
        # remove the chunked collection so future tests will always chunk
        client_manager = ClientManager()
        with client_manager.connect_to_client() as client:
            if client.collections.exists("ELYSIA_CHUNKED_weaviate_documentation__"):
                client.collections.delete("ELYSIA_CHUNKED_weaviate_documentation__")
            if client.collections.exists("ELYSIA_CHUNKED_weaviate_blogs__"):
                client.collections.delete("ELYSIA_CHUNKED_weaviate_blogs__")


@pytest.mark.asyncio
async def test_impossible_query():
    """
    Test that when given a task that is not possible, the model does not attempt it (too many times).
    """

    # if the collection is not found, skip the test
    client_manager = ClientManager()
    if not client_manager.client.collections.exists("Weather"):
        pytest.skip("Collection not found in Weaviate")

    user_prompt = "What is the weather in January 2025?"
    tree = await run_tree(user_prompt, ["Weather"])

    metric = metrics.GEval(
        name="Judge of agent knowledge lacking awareness",
        criteria="""
        The model should not attempt to answer a question that is not possible to answer.
        So the model may attempt to answer this but should eventually realise it is not possible.
        """,
        evaluation_params=[
            LLMTestCaseParams.INPUT,
            LLMTestCaseParams.ACTUAL_OUTPUT,
            LLMTestCaseParams.TOOLS_CALLED,
            LLMTestCaseParams.EXPECTED_OUTPUT,
        ],
    )

    test_case = LLMTestCase(
        input=user_prompt,
        actual_output=tree.tree_data.conversation_history[-1]["content"],
        tools_called=get_tools_called(tree, user_prompt),
        expected_output="""
        January 2025 is not a valid date in the weather dataset, only 2014-2016 is available.
        So the model should eventually realise the data is not available.
        """,
    )

    res = evaluate(test_cases=[test_case], metrics=[metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason



import pytest

from fastapi.responses import JSONResponse

import json
import os
import random
from elysia.api.api_types import ProcessCollectionData, SaveConfigUserData
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.init import initialise_user
from elysia.api.routes.processor import process_collection
from elysia.api.routes.collections import collection_metadata
from elysia.util.client import ClientManager
from elysia.api.core.log import logger, set_log_level
from elysia.preprocessing.collection import (
    delete_preprocessed_collection_async,
    preprocessed_collection_exists_async,
)
from elysia.api.routes.user_config import get_current_user_config, save_config_user
from elysia.api.routes.collections import update_metadata, delete_metadata
from weaviate.classes.config import Configure
from weaviate.classes.query import Filter
from weaviate.util import generate_uuid5
import weaviate.classes as wvc

from elysia.api.api_types import (
    ViewPaginatedCollectionData,
    UpdateCollectionMetadataData,
    MetadataNamedVectorData,
    MetadataFieldData,
    InitialiseTreeData,
)

set_log_level("CRITICAL")


def create_named_vectors_collection(
    client_manager: ClientManager, collection_name: str
):

    with client_manager.connect_to_client() as client:

        # Delete collection if it exists
        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)

        # Create collection with multiple named vectors
        collection = client.collections.create(
            name=collection_name,
            properties=[
                wvc.config.Property(
                    name="product_id", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(
                    name="category", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(name="author", data_type=wvc.config.DataType.TEXT),
            ],
            vector_config=[
                wvc.config.Configure.Vectors.text2vec_openai(
                    name="content_vector", source_properties=["content"]
                ),
                wvc.config.Configure.Vectors.text2vec_openai(
                    name="title_vector", source_properties=["title"]
                ),
                wvc.config.Configure.Vectors.self_provided(name="custom_vector"),
            ],
        )

        # Toy data for named vectors collection
        documents_data = [
            {
                "product_id": "doc1",
                "title": "Introduction to Machine Learning",
                "content": "Machine learning is a subset of artificial intelligence that focuses on algorithms that can learn from data.",
                "category": "Education",
                "author": "Dr. Smith",
                "custom_vector": [0.1, 0.2, 0.3, 0.4, 0.5],  # Custom embedding
            },
            {
                "product_id": "doc2",
                "title": "Python Programming Basics",
                "content": "Python is a versatile programming language used in web development, data science, and automation.",
                "category": "Programming",
                "author": "Jane Doe",
                "custom_vector": [0.6, 0.7, 0.8, 0.9, 1.0],
            },
            {
                "product_id": "doc3",
                "title": "Database Design Principles",
                "content": "Good database design involves normalization, indexing, and understanding relationships between entities.",
                "category": "Database",
                "author": "Bob Johnson",
                "custom_vector": [0.2, 0.4, 0.6, 0.8, 0.3],
            },
        ]

        # Import data with named vectors
        with collection.batch.fixed_size(batch_size=10) as batch:
            for doc in documents_data:
                batch.add_object(
                    properties={
                        "title": doc["title"],
                        "content": doc["content"],
                        "category": doc["category"],
                        "author": doc["author"],
                    },
                    vector={"custom_vector": doc["custom_vector"]},
                    uuid=generate_uuid5(doc["product_id"]),
                )


def create_regular_vectorizer_collection(
    client_manager: ClientManager, collection_name: str
):

    with client_manager.connect_to_client() as client:

        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)

        collection = client.collections.create(
            name=collection_name,
            properties=[
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(
                    name="description", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(
                    name="tags", data_type=wvc.config.DataType.TEXT_ARRAY
                ),
                wvc.config.Property(name="price", data_type=wvc.config.DataType.NUMBER),
            ],
            vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_openai(
                vectorize_collection_name=False
            ),
        )

        products_data = [
            {
                "product_id": "prod1",
                "title": "Wireless Bluetooth Headphones",
                "description": "High-quality wireless headphones with noise cancellation and 30-hour battery life.",
                "tags": ["electronics", "audio", "wireless", "bluetooth"],
                "price": 199.99,
            },
            {
                "product_id": "prod2",
                "title": "Organic Green Tea",
                "description": "Premium organic green tea sourced from high-altitude tea gardens in Japan.",
                "tags": ["beverage", "tea", "organic", "healthy"],
                "price": 24.99,
            },
            {
                "product_id": "prod3",
                "title": "Yoga Mat with Alignment Lines",
                "description": "Non-slip yoga mat with helpful alignment lines for proper posture during practice.",
                "tags": ["fitness", "yoga", "exercise", "wellness"],
                "price": 39.99,
            },
            {
                "product_id": "prod4",
                "title": "Smart Home Security Camera",
                "description": "WiFi-enabled security camera with motion detection and smartphone notifications.",
                "tags": ["security", "smart-home", "camera", "surveillance"],
                "price": 149.99,
            },
        ]

        with collection.batch.fixed_size(batch_size=10) as batch:
            for product in products_data:
                batch.add_object(
                    properties={
                        "title": product["title"],
                        "description": product["description"],
                        "tags": product["tags"],
                        "price": product["price"],
                    },
                    uuid=generate_uuid5(product["product_id"]),
                )


user_manager = get_user_manager()


def read_response(response: JSONResponse):
    return json.loads(response.body)


class fake_websocket:
    def __init__(self):
        self.results = []

    async def send_json(self, data: dict):
        self.results.append(data)


@pytest.mark.asyncio
async def test_process_no_api_keys():
    """
    Test processing a collection with no api keys.
    """

    user_id = "test_user_Process_no_api_keys"
    collection_name = "Test_ELYSIA_collection_process_no_api_keys"

    await initialise_user(
        user_id,
        user_manager,
    )
    local_user = await user_manager.get_user_local(user_id)
    client_manager = ClientManager()
    create_regular_vectorizer_collection(client_manager, collection_name)

    try:
        # get config
        response = await get_current_user_config(user_id, user_manager)
        response = read_response(response)
        config_id = response["config"]["id"]
        config_name = response["config"]["name"]

        # remove api keys
        await save_config_user(
            user_id,
            config_id,
            SaveConfigUserData(
                name=config_name,
                default=True,
                config={
                    "settings": {
                        "BASE_MODEL": "gemini-2.0-flash-001",
                        "BASE_PROVIDER": "openrouter/google",
                        "COMPLEX_MODEL": "gemini-2.0-flash-001",
                        "COMPLEX_PROVIDER": "openrouter/google",
                        "WCD_URL": os.getenv("WCD_URL"),
                        "WCD_API_KEY": os.getenv("WCD_API_KEY"),
                        "API_KEYS": {},
                    }
                },
                frontend_config={},
            ),
            user_manager,
        )

        websocket = fake_websocket()
        await process_collection(
            ProcessCollectionData(
                user_id=user_id, collection_name=collection_name
            ).model_dump(),
            websocket,
            user_manager,
        )

        error_found = False
        for result in websocket.results:
            error_found = result["error"] != ""
        assert error_found

    finally:

        # delete the collections
        try:
            with client_manager.connect_to_client() as client:

                if client.collections.exists(collection_name):
                    client.collections.delete(collection_name)

                if await preprocessed_collection_exists_async(
                    collection_name, client_manager
                ):
                    await delete_preprocessed_collection_async(
                        collection_name, client_manager
                    )

        except Exception as e:
            print(f"Can't delete collection: {str(e)}")
            pass


@pytest.mark.asyncio
async def test_process_collection():

    user_id = "test_user_Process_collection"
    collection_name = "Test_ELYSIA_collection_process_collection"

    client_manager = ClientManager()
    create_regular_vectorizer_collection(client_manager, collection_name)
    try:
        await initialise_user(
            user_id,
            user_manager,
        )
        local_user = await user_manager.get_user_local(user_id)
        client_manager = local_user["client_manager"]

        websocket = fake_websocket()
        await process_collection(
            ProcessCollectionData(
                user_id=user_id, collection_name=collection_name
            ).model_dump(),
            websocket,
            user_manager,
        )
        for i, result in enumerate(websocket.results):
            if result["error"] != "":
                assert (
                    False
                ), f"({i}) Message i-1: {websocket.results[i-1]['message']}, error: {result['error']}"

        for result in websocket.results[:-1]:
            assert result["type"] == "update"
            assert result["progress"] < 1

        assert websocket.results[-1]["type"] == "completed"
        assert websocket.results[-1]["progress"] == 1

        with client_manager.connect_to_client() as client:
            assert client.collections.exists(collection_name)
            assert client.collections.exists("ELYSIA_METADATA__")
            metadata_collection = client.collections.get("ELYSIA_METADATA__")
            metadata = metadata_collection.query.fetch_objects(
                filters=Filter.by_property("name").equal(collection_name),
                limit=1,
            )
            assert len(metadata.objects) >= 1

    finally:

        # delete the collections
        try:
            with client_manager.connect_to_client() as client:

                if client.collections.exists(collection_name):
                    client.collections.delete(collection_name)

                if await preprocessed_collection_exists_async(
                    collection_name, client_manager
                ):
                    await delete_preprocessed_collection_async(
                        collection_name, client_manager
                    )

        except Exception as e:
            print(f"Can't delete collection: {str(e)}")
            pass


@pytest.mark.asyncio
async def test_collection_metadata():

    user_id = "test_user_collection_metadata"
    conversation_id = "test_conversation_collection_metadata"
    collection_name = "Test_ELYSIA_collection_collection_metadata"

    client_manager = ClientManager()
    user_manager = get_user_manager()
    create_named_vectors_collection(client_manager, collection_name)

    await initialise_user(
        user_id,
        user_manager,
    )

    try:

        websocket = fake_websocket()
        await process_collection(
            ProcessCollectionData(
                user_id=user_id, collection_name=collection_name
            ).model_dump(),
            websocket,
            user_manager,
        )

        basic = await collection_metadata(
            user_id,
            collection_name=collection_name,
            user_manager=user_manager,
        )
        basic = read_response(basic)
        assert basic["error"] == "", basic["error"]

        assert "summary" in basic["metadata"]
        assert "fields" in basic["metadata"]
        assert "mappings" in basic["metadata"]
        assert "length" in basic["metadata"]
        assert "mappings" in basic["metadata"]
        assert "index_properties" in basic["metadata"]
        assert "named_vectors" in basic["metadata"]
        assert basic["metadata"]["named_vectors"] is not None
        assert len(basic["metadata"]["named_vectors"]) == 3
        assert all(nv["enabled"] for nv in basic["metadata"]["named_vectors"])

        named_vector_names = [nv["name"] for nv in basic["metadata"]["named_vectors"]]
        assert "content_vector" in named_vector_names
        assert "title_vector" in named_vector_names
        assert "custom_vector" in named_vector_names

        assert "vectorizer" in basic["metadata"]
        assert basic["metadata"]["vectorizer"] is None

    finally:
        async with client_manager.connect_to_async_client() as client:
            if await client.collections.exists(collection_name):
                await client.collections.delete(collection_name)


@pytest.mark.asyncio
async def test_update_metadata():

    user_id = "test_user_update_metadata"
    conversation_id = "test_conversation_update_metadata"

    collection_name = "Test_ELYSIA_collection_update_metadata"
    client_manager = ClientManager()
    user_manager = get_user_manager()
    create_named_vectors_collection(client_manager, collection_name)

    try:
        await initialise_user(
            user_id,
            user_manager,
        )

        # preprocess
        websocket = fake_websocket()
        await process_collection(
            ProcessCollectionData(
                user_id=user_id, collection_name=collection_name
            ).model_dump(),
            websocket,
            user_manager,
        )

        # retrieve the metadata
        metadata = await collection_metadata(
            user_id=user_id,
            collection_name=collection_name,
            user_manager=user_manager,
        )
        metadata = read_response(metadata)
        assert metadata["error"] == ""
        metadata = metadata["metadata"]

        # record values
        old_named_vectors = metadata["named_vectors"]
        old_summary = metadata["summary"]
        old_mappings = metadata["mappings"]
        old_fields = metadata["fields"]

        # what to update
        new_named_vectors = [
            MetadataNamedVectorData(
                name="content_vector",
                enabled=True,
                description="Test description for content vector.",
            )
        ]

        new_summary = "This is a test summary."
        new_mappings = {
            "product": {
                "name": "title",
                "category": "content",
                "description": "author",
            }
        }
        new_fields = [
            MetadataFieldData(
                name="title",
                description="A test description for title.",
            )
        ]

        # update the metadata
        update_metadata_response = await update_metadata(
            user_id=user_id,
            collection_name=collection_name,
            data=UpdateCollectionMetadataData(
                named_vectors=new_named_vectors,
                summary=new_summary,
                mappings=new_mappings,
                fields=new_fields,
            ),
            user_manager=user_manager,
        )
        update_metadata_response = read_response(update_metadata_response)
        assert update_metadata_response["error"] == "", update_metadata_response[
            "error"
        ]

        # retrieve the metadata again
        metadata = await collection_metadata(
            user_id=user_id,
            collection_name=collection_name,
            user_manager=user_manager,
        )
        metadata = read_response(metadata)
        assert metadata["error"] == ""
        metadata = metadata["metadata"]

        # check the values have been updated
        for named_vector in metadata["named_vectors"]:
            if named_vector["name"] == "content_vector":
                assert named_vector["enabled"] == True
                assert (
                    named_vector["description"]
                    == "Test description for content vector."
                )
                break

        assert metadata["summary"] == new_summary
        assert "product" in metadata["mappings"]
        for field in metadata["fields"]:
            if field["name"] == "title":
                assert field["description"] == new_fields[0].description
                break

        # check the other named vectors have not changed
        for named_vector in metadata["named_vectors"]:
            if named_vector["name"] != "content_vector":
                assert named_vector["enabled"]
                assert named_vector["description"] == ""
    finally:
        async with client_manager.connect_to_async_client() as client:
            if await client.collections.exists(collection_name):
                await client.collections.delete(collection_name)


@pytest.mark.asyncio
async def test_delete_metadata():
    user_id = "test_user_delete_metadata"
    conversation_id = "test_conversation_delete_metadata"
    collection_name = "Test_ELYSIA_collection_delete_metadata"

    user_manager = get_user_manager()
    client_manager = ClientManager()

    await initialise_user(user_id, user_manager)

    # create a collection
    create_named_vectors_collection(client_manager, collection_name)

    # preprocess
    websocket = fake_websocket()
    await process_collection(
        ProcessCollectionData(
            user_id=user_id, collection_name=collection_name
        ).model_dump(),
        websocket,
        user_manager,
    )
    for i, result in enumerate(websocket.results):
        if result["error"] != "":
            assert (
                False
            ), f"({i}) Message i-1: {websocket.results[i-1]['message']}, error: {result['error']}"

    # check the metadata has been created
    metadata = await collection_metadata(
        user_id=user_id,
        collection_name=collection_name,
        user_manager=user_manager,
    )
    metadata = read_response(metadata)
    assert metadata["error"] == "", metadata["error"]
    assert metadata["metadata"] is not {} and metadata["metadata"] is not None

    # delete the metadata
    delete_metadata_response = await delete_metadata(
        user_id=user_id,
        collection_name=collection_name,
        user_manager=user_manager,
    )
    delete_metadata_response = read_response(delete_metadata_response)
    assert delete_metadata_response["error"] == "", delete_metadata_response["error"]

    # check that the metadata has been deleted
    metadata = await collection_metadata(
        user_id=user_id,
        collection_name=collection_name,
        user_manager=user_manager,
    )
    metadata = read_response(metadata)
    assert metadata["error"] != ""



import dspy
from typing import List
from elysia.tools.text.objects import TextWithCitation


class CitedSummarizingPrompt(dspy.Signature):
    """
    Given a user_prompt, as well as a list of retrieved objects from the environment, summarize the information in the objects to answer the user's prompt.
    Your output will be a collection of TextWithCitation objects that form a complete summary when combined.
    Do not list any of the retrieved objects in your response. Do not give an itemised list of the objects, since they will be displayed to the user anyway.
    Your summary should be parsing the retrieved objects, and summarising the information in them that is relevant to the user's prompt.
    For example, picking the most relevant information from the retrieved objects, and discussing it, repeating ONLY relevant information if necessary.
    You should provide useful analysis, new information via analysing the existing objects, and synthesising the information.
    """

    subtitle = dspy.OutputField(description="A subtitle for the summary")
    cited_text: List[TextWithCitation] = dspy.OutputField(
        description="""
        A list of TextWithCitation objects whose 'text' fields will be concatenated to form the complete summary.
        
        FORMATTING RULES:
        1. Use markdown for structure (headings, lists, etc.), this is encouraged and recommended to break down the information to be more readable.
        2. IMPORTANT: Always add newline characters (\\n) before and after markdown elements (lists, headers)
        3. When a formatting element spans multiple TextWithCitation objects, ensure each object has proper newlines
        4. Example: If a bullet point list spans multiple objects, each object must end with \\n and the next begin with proper indentation
        
        CITATION RULES:
        1. NEVER include reference IDs in the 'text' field itself
        2. NEVER include the literal string "_REF_ID" in your text
        3. NEVER create hyperlinks to references in the text field
        4. ALWAYS copy reference IDs exactly from the source's _REF_ID to the ref_ids list
        5. Keep the text clean and reference-free: all citations happen only through the ref_ids list
        6. Since you are referencing via _REF_ID, you do not need any other type of referencing such as URLs or hyperlinks.

        CORRECT EXAMPLE: text: "Here is some text", ref_ids: ["example_citation_1", "example_citation_2"]
        INCORRECT EXAMPLE: text: "Here is some text _REF_ID", ref_ids: ["example_citation_1", "example_citation_2"]
        INCORRECT EXAMPLE: text: "Here is some text [example_citation_1]", ref_ids: ["example_citation_1", "example_citation_2"]
        
        CONTENT GUIDELINES:
        1. Focus only on information relevant to the user's prompt
        2. Provide analysis and synthesis, not just extraction
        3. Only cite information that exists in the retrieved objects
        4. Each text segment should map logically to its citations
        5. Use multiple TextWithCitation objects to align text with its specific sources
        6. Do not just repeat information from the environment. Create insights, offer suggestions, do not just list objects.
        """.strip()
    )


class SummarizingPrompt(dspy.Signature):
    """
    Given a user_prompt, as well as a list of retrieved objects, summarize the information in the objects to answer the user's prompt.
    Information about you:
    - You are a chatbot for an app named Elysia.
    - You are a helpful assistant designed to be used in a chat interface and respond to user's prompts in a helpful, friendly, and polite manner.
    - Your primary task is to summarize the information in the retrieved objects to answer the user's prompt.
    Do not list any of the retrieved objects in your response. Do not give an itemised list of the objects, since they will be displayed to the user anyway.
    """

    subtitle = dspy.OutputField(description="A subtitle for the summary")
    summary = dspy.OutputField(
        description="""
        The summary of the retrieved objects. You can use markdown formatting.
        Don't provide an itemised list of the objects, since they will be displayed to the user anyway.
        Your summary should take account what the user prompt is, and the information in the retrieved objects.
        DO NOT list the retrieved objects in your summary.
        Your summary should be parsing the retrieved objects, and summarising the information in them that is relevant to the user's prompt.
        For example, picking the most relevant information from the retrieved objects, and discussing it, repeating ONLY relevant information if necessary.
        """.strip()
    )


class TextResponsePrompt(dspy.Signature):
    """
    You are a helpful assistant, designed to be used in a chat interface and respond to user's prompts in a helpful, friendly, and polite manner.
    Given a user_prompt, as well as a list of retrieved objects, respond to the user's prompt.
    Your response should be informal, polite, and assistant-like.
    Information about you:
    - You are a chatbot for an app named Elysia.
    - You are a helpful assistant designed to be used in a chat interface and respond to user's prompts in a helpful, friendly, and polite manner.
    - Your primary task is to respond to the user's query.
    Do not list any of the retrieved objects in your response. Do not give an itemised list of the objects, since they will be displayed to the user anyway.
    If there is an error and you could not complete a task, use this tool to suggest a brief reason why.
    If, for example, there is a missing API key, then the user needs to add it to the settings (which you should inform them of).
    Or you cannot connect to weaviate, then the user needs to input their API keys in the settings.
    If there are no collections available, the user needs to analyze this in the 'data' tab.
    If there are other problems, and it looks like the user can fix it, then provide a suggestion.
    """

    response = dspy.OutputField(
        description="""
        The response to the user's prompt.
        If you are explaining how something went wrong, or you could not complete a task, suggest a brief reason why.
        You can suggest alternative questions for the user to ask, alternative ways of phrasing the prompt, or explain your own limitations and how to avoid them.
        E.g., you could ask the user to be more specific, or to provide more information, or explain that you are limited by your data. Be creative and adaptive based on the user's prompt and the type of issue that occurred.
        Use present tense in your text, as if you are currently completing the action.
        Use gender neutral language.
        """.strip()
    )



import pytest
import inspect

from elysia.tools.retrieval.chunk import AsyncCollectionChunker, Chunker
from elysia.util.client import ClientManager

from weaviate.classes.query import QueryReference
from weaviate.classes.config import Configure, DataType, Property, ReferenceProperty
from weaviate.collections.classes.data import DataObject, DataReference
from weaviate.collections import CollectionAsync
from weaviate.collections.classes.internal import Object, QueryReturn
from weaviate.collections.classes.config_vectorizers import _Vectorizer
from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.util import generate_uuid5
from weaviate.client import WeaviateAsyncClient


@pytest.mark.asyncio
async def test_correct_vectoriser():
    collection_name_full = "Test_ELYSIA_collection_chunking_vectoriser_full"
    collection_name_named = "Test_ELYSIA_collection_chunking_vectoriser_named"
    client_manager = ClientManager()

    # example data
    data = [
        {
            "content": (
                "Lorem ipsum dolor sit amet consectetur adipiscing elit. "
                "Quisque faucibus ex sapien vitae pellentesque sem placerat. "
                "In id cursus mi pretium tellus duis convallis. "
                "Tempus leo eu aenean sed diam urna tempor. "
                "Pulvinar vivamus fringilla lacus nec metus bibendum egestas. "
                "Iaculis massa nisl malesuada lacinia integer nunc posuere. "
                "Ut hendrerit semper vel class aptent taciti sociosqu. "
                "Ad litora torquent per conubia nostra inceptos himenaeos."
            ),
            "other_field": "other_value",
        },
        {
            "content": (
                "Lorem ipsum dolor sit sit amet consectetur adipiscing elit. "
                "Quisque faucibus ex sapien sed vitae pellentesque sem placerat. "
                "In id cursus mi pretium tellus sed duis convallis. "
                "Tempus leo eu aenean sed diam urna tempor. "
                "Pulvinar vivamus fringilla la cus nec metus bibendum egestas. "
                "Iaculis massa nisl malesuada   lacinia integer nunc posuere. "
                "Ut hendrerit semper vel class ooga aptent taciti sociosqu. "
                "Ad litora torquent per conubia booga nostra inceptos himenaeos."
            ),
            "other_field": "other_value_2",
        },
    ]

    # create collection with vectoriser
    full_vectoriser = Configure.Vectorizer.text2vec_openai(
        model="text-embedding-3-large",
        dimensions=256,
    )
    async with client_manager.connect_to_async_client() as client:
        collection_full = await client.collections.create(
            collection_name_full,
            vectorizer_config=full_vectoriser,
        )
        await collection_full.data.insert_many(data)

    # create collection with named vectoriser
    named_vectoriser = Configure.NamedVectors.text2vec_jinaai(
        name="content_vector", source_properties=["content"], model="jina-embeddings-v3"
    )
    async with client_manager.connect_to_async_client() as client:
        collection_named = await client.collections.create(
            collection_name_named,
            vectorizer_config=[
                named_vectoriser,
            ],
        )
        await collection_named.data.insert_many(data)

    try:
        # do chunking on full collection
        collection_chunker = AsyncCollectionChunker(collection_name_full)
        await collection_chunker.create_chunked_reference("content", client_manager)

        async with client_manager.connect_to_async_client() as client:
            # check existence of collection
            assert await client.collections.exists(collection_name_full)

            # check vectoriser
            collection_full_config = await client.collections.get(
                collection_name_full
            ).config.get()
            assert (
                collection_full_config.vectorizer_config.vectorizer
                == full_vectoriser.vectorizer
            )

        # check named vectoriser
        collection_chunker = AsyncCollectionChunker(collection_name_named)
        await collection_chunker.create_chunked_reference("content", client_manager)

        async with client_manager.connect_to_async_client() as client:
            # check existence of collection
            assert await client.collections.exists(collection_name_named)

            # check vectoriser
            collection_named_config = await client.collections.get(
                collection_name_named
            ).config.get()
            assert collection_named_config.vector_config[
                list(collection_named_config.vector_config.keys())[0]
            ].vectorizer.source_properties == ["content"]
            assert (
                collection_named_config.vector_config[
                    list(collection_named_config.vector_config.keys())[0]
                ].vectorizer.vectorizer
                == named_vectoriser.vectorizer.vectorizer
            )

    finally:
        async with client_manager.connect_to_async_client() as client:
            await client.collections.delete(collection_name_full)
            await client.collections.delete(collection_name_named)

        await client_manager.close_clients()



import ast
import datetime
from typing import Any, List

from weaviate.classes.config import DataType
from weaviate.classes.query import Filter, Sort
from weaviate.types import UUID

from elysia.util.parsing import format_datetime

data_mapping = {
    "text": DataType.TEXT,
    "int": DataType.INT,
    "number": DataType.NUMBER,
    "bool": DataType.BOOL,
    "date": DataType.DATE,
    "object": DataType.OBJECT,
    "text[]": DataType.TEXT_ARRAY,
    "int[]": DataType.INT_ARRAY,
    "number[]": DataType.NUMBER_ARRAY,
    "bool[]": DataType.BOOL_ARRAY,
    "date[]": DataType.DATE_ARRAY,
    "object[]": DataType.OBJECT_ARRAY,
}


async def retrieve_all_collection_names(client):
    all_collections = await client.collections.list_all()
    return [
        collection
        for collection in all_collections
        if not collection.startswith("ELYSIA_")
    ]


def get_collection_data_types(client, collection_name: str):
    properties = client.collections.get(collection_name).config.get().properties
    return {property.name: property.data_type[:] for property in properties}


async def async_get_collection_data_types(client, collection_name: str):
    config = client.collections.get(collection_name).config
    config = await config.get()
    properties = config.properties
    return {property.name: property.data_type[:] for property in properties}


def get_collection_weaviate_data_types(client, collection_name: str):
    data_types = get_collection_data_types(client, collection_name)
    return {k: data_mapping[v] for k, v in data_types.items()}


async def async_get_collection_weaviate_data_types(client, collection_name: str):
    data_types = await async_get_collection_data_types(client, collection_name)
    return {k: data_mapping[v] for k, v in data_types.items()}


def convert_weaviate_list(list_object: list):
    for i, value in enumerate(list_object):
        if isinstance(value, str) and value.startswith("[") and value.endswith("]"):
            list_object[i] = ast.literal_eval(value)

        if isinstance(value, dict):
            list_object[i] = convert_weaviate_object(value)
        elif isinstance(value, list):
            list_object[i] = convert_weaviate_list(value)
        elif isinstance(value, datetime.datetime):
            list_object[i] = format_datetime(value)
        elif isinstance(value, UUID):
            list_object[i] = str(value)
        elif (
            not isinstance(value, str)
            and not isinstance(value, list)
            and not isinstance(value, dict)
            and not isinstance(value, float)
            and not isinstance(value, int)
            and not isinstance(value, bool)
        ):
            list_object[i] = str(value)

    return list_object


def convert_weaviate_object(dict_object: dict):
    for key, value in dict_object.items():

        if isinstance(value, datetime.datetime):
            dict_object[key] = format_datetime(value)
        elif isinstance(value, list):
            dict_object[key] = convert_weaviate_list(value)
        elif isinstance(value, dict):
            dict_object[key] = convert_weaviate_object(value)
        elif isinstance(value, UUID):
            dict_object[key] = str(value)
        elif (
            not isinstance(value, str)
            and not isinstance(value, list)
            and not isinstance(value, dict)
            and not isinstance(value, float)
            and not isinstance(value, int)
            and not isinstance(value, bool)
        ):
            dict_object[key] = str(value)

    return dict_object


async def paginated_collection(
    client,
    collection_name: str,
    query: str = "",
    sort_on: str | None = None,
    ascending: bool = False,
    filter_config: dict[
        str, Any
    ] = {},  # e.g. {"type":"all", "filters": [{"field":"issue_state", "operator":"not_equal", "value":"open"}]}
    page_size: int = 10,
    page_number: int = 1,
):
    collection = client.collections.get(collection_name)

    filter_type = filter_config.get("type", "all")
    filters_list = filter_config.get("filters", [])
    filters = [f["field"] for f in filters_list]
    filter_operators = [f["operator"] for f in filters_list]
    filter_values = [f["value"] for f in filters_list]

    if len(filters) == 0:

        if query != "":
            response = await collection.query.bm25(
                query=query,
                limit=page_size,
                offset=page_size * (page_number - 1),
            )
        elif sort_on is not None:
            response = await collection.query.fetch_objects(
                sort=Sort.by_property(name=sort_on, ascending=ascending),
                limit=page_size,
                offset=page_size * (page_number - 1),
            )
        else:
            response = await collection.query.fetch_objects(
                limit=page_size, offset=page_size * (page_number - 1)
            )

    else:
        all_filters = []

        for filter_name, filter_operator, filter_value in zip(
            filters, filter_operators, filter_values
        ):
            if filter_operator == "equal":
                all_filters.append(Filter.by_property(filter_name).equal(filter_value))
            elif filter_operator == "greater_or_equal":
                all_filters.append(
                    Filter.by_property(filter_name).greater_or_equal(filter_value)
                )
            elif filter_operator == "greater_than":
                all_filters.append(
                    Filter.by_property(filter_name).greater_than(filter_value)
                )
            elif filter_operator == "less_or_equal":
                all_filters.append(
                    Filter.by_property(filter_name).less_or_equal(filter_value)
                )
            elif filter_operator == "less_than":
                all_filters.append(
                    Filter.by_property(filter_name).less_than(filter_value)
                )
            elif filter_operator == "not_equal":
                all_filters.append(
                    Filter.by_property(filter_name).not_equal(filter_value)
                )
            else:
                raise ValueError(f"Invalid filter operator: {filter_operator}")

        if filter_type == "all":
            main_filter = Filter.all_of(all_filters)
        elif filter_type == "any":
            main_filter = Filter.any_of(all_filters)
        else:
            raise ValueError(f"Invalid filter type: {filter_type}")

        if query != "":
            response = await collection.query.bm25(
                query=query,
                filters=main_filter,
                limit=page_size,
                offset=page_size * (page_number - 1),
            )
        elif sort_on is not None:
            response = await collection.query.fetch_objects(
                filters=main_filter,
                sort=Sort.by_property(name=sort_on, ascending=ascending),
                limit=page_size,
                offset=page_size * (page_number - 1),
            )
        else:
            response = await collection.query.fetch_objects(
                filters=main_filter,
                limit=page_size,
                offset=page_size * (page_number - 1),
            )

    objects = [
        {
            **convert_weaviate_object(o.properties),
            "uuid": str(o.uuid),
        }
        for o in response.objects
    ]

    return objects



import os

import torch
import transformers
from deepeval.models import DeepEvalBaseLLM
from dotenv import load_dotenv
from transformers import AutoModelForCausalLM, AutoTokenizer

load_dotenv(override=True)


class CustomLlama(DeepEvalBaseLLM):
    def __init__(self, max_length: int = 5000):
        self.max_length = max_length

        model_4bit = AutoModelForCausalLM.from_pretrained(
            "meta-llama/Llama-3.2-3B-Instruct",
            device_map="auto",
            token=os.getenv("HF_TOKEN"),
        )
        tokenizer = AutoTokenizer.from_pretrained(
            "meta-llama/Llama-3.2-3B-Instruct",
            token=os.getenv("HF_TOKEN"),
        )

        self.model = model_4bit
        self.tokenizer = tokenizer

    def load_model(self):
        return self.model

    def generate(self, prompt: str) -> str:
        model = self.load_model()

        pipeline = transformers.pipeline(
            "text-generation",
            model=model,
            tokenizer=self.tokenizer,
            use_cache=True,
            device_map="auto",
            max_length=self.max_length,
            do_sample=True,
            top_k=5,
            num_return_sequences=1,
            eos_token_id=self.tokenizer.eos_token_id,
            pad_token_id=self.tokenizer.eos_token_id,
        )

        return pipeline(prompt)

    async def a_generate(self, prompt: str) -> str:
        return self.generate(prompt)

    def get_model_name(self):
        return "Llama-3 8B"


# if __name__ == "__main__":
#     llm = CustomLlama()
#     print(llm.generate("Hello, how are you?"))

#     from deepeval.metrics import AnswerRelevancyMetric

#     metric = AnswerRelevancyMetric(model=llm)
#     metric.measure(
#         model=llm,
#         data=data,
#     )



import pytest
from deepeval import evaluate, metrics
from deepeval.test_case import LLMTestCase, LLMTestCaseParams

from elysia import Tree, Settings, tool


def test_tree_no_client():

    settings = Settings.from_smart_setup()
    settings.WCD_URL = ""
    settings.WCD_API_KEY = ""

    tree = Tree(
        low_memory=False,
        branch_initialisation="one_branch",
        settings=settings,
    )

    response, objects = tree.run(
        "What was the most recent message in the messages collection?",
        collection_names=["Test_ELYSIA_Messages_collection"],
    )

    metric = metrics.GEval(
        name="Judge of agent being unable to connect to Weaviate",
        criteria="""
        The response should inform the user that the agent is unable to connect to Weaviate,
        and to set the WCD_URL and WCD_API_KEY in the settings.
        OR, the response should inform the user to analyse the data.
        """,
        evaluation_params=[
            LLMTestCaseParams.INPUT,
            LLMTestCaseParams.ACTUAL_OUTPUT,
            LLMTestCaseParams.EXPECTED_OUTPUT,
        ],
    )

    test_case = LLMTestCase(
        input="What was the most recent message in the messages collection?",
        actual_output=response,
        expected_output="""
        The agent is unable to connect to Weaviate,
        and to set the WCD_URL and WCD_API_KEY in the settings.
        OR, the response should inform the user to analyse the data.
        """,
    )

    res = evaluate(test_cases=[test_case], metrics=[metric])
    for test_case in res.test_results:
        assert test_case.success, test_case.metrics_data[0].reason


@pytest.mark.asyncio
async def test_tool_decorator():
    @tool
    async def add_two_numbers(x: int, y: int) -> int:
        return x + y

    tree = Tree(
        low_memory=False,
        branch_initialisation="one_branch",
        settings=Settings.from_smart_setup(),
    )
    tree.add_tool(add_two_numbers)

    assert "add_two_numbers" in tree.tools

    prompt = "What is the sum of 1123213 and 2328942390843209?"
    response, objects = tree.run(prompt)

    assert "2328942391966422" in response

    found = False
    for action in tree.actions_called[prompt]:
        if (
            1123213 in action["inputs"].values()
            and 2328942390843209 in action["inputs"].values()
        ):
            found = True
            break

    assert found


@pytest.mark.asyncio
async def test_tool_decorator_with_args():

    tree = Tree(
        low_memory=False,
        branch_initialisation="one_branch",
        settings=Settings.from_smart_setup(),
    )

    @tool(tree=tree)
    async def add_two_numbers(x: int, y: int) -> int:
        return x + y

    assert "add_two_numbers" in tree.tools

    prompt = "What is the sum of 1123213 and 2328942390843209?"
    response, objects = tree.run(prompt)

    assert "2328942391966422" in response

    found = False
    for action in tree.actions_called[prompt]:
        if (
            1123213 in action["inputs"].values()
            and 2328942390843209 in action["inputs"].values()
        ):
            found = True
            break

    assert found

    tree = Tree()

    @tool(tree=tree, status="Adding two numbers", end=True)
    async def add_two_numbers2(x: int, y: int) -> int:
        """
        This function adds two numbers together.
        """
        return x + y



# FastAPI
from fastapi import APIRouter, Depends
from fastapi.responses import JSONResponse

from elysia.api.api_types import (
    DebugData,
    FollowUpSuggestionsData,
    NERData,
    TitleData,
)

from elysia.tree.tree import Tree

# Logging
from elysia.api.core.log import logger

# Dependencies
from elysia.api.dependencies.common import get_user_manager

# Services
from elysia.api.services.user import UserManager

# Settings
from elysia.config import nlp

# util
from elysia.api.core.log import logger

router = APIRouter()


@router.post("/ner")
async def named_entity_recognition(data: NERData):
    """
    Performs Named Entity Recognition using spaCy.
    Returns a list of entities with their labels, start and end positions.
    """
    logger.debug(f"/ner API request received")
    logger.debug(f"Text: {data.text}")

    try:

        doc = nlp(data.text)
        out = {"text": data.text, "entity_spans": [], "noun_spans": [], "error": ""}

        for ent in doc.ents:
            out["entity_spans"].append((ent.start_char, ent.end_char))

        # Get noun spans
        for token in doc:
            if token.pos_ == "NOUN":
                span = doc[token.i : token.i + 1]
                out["noun_spans"].append((span.start_char, span.end_char))

        return JSONResponse(content=out, status_code=200)

    except Exception as e:
        logger.exception(f"Error in /ner API")
        return JSONResponse(
            content={
                "text": data.text,
                "entity_spans": [],
                "noun_spans": [],
                "error": str(e),
            },
            status_code=200,
        )


@router.post("/title")
async def title(data: TitleData, user_manager: UserManager = Depends(get_user_manager)):
    logger.debug(f"/title API request received")
    logger.debug(f"User ID: {data.user_id}")
    logger.debug(f"Conversation ID: {data.conversation_id}")
    logger.debug(f"Text: {data.text}")

    if user_manager.check_tree_timeout(data.user_id, data.conversation_id):
        logger.warning(
            f"(/title) Conversation {data.conversation_id} has timed out for user {data.user_id}"
        )
        return JSONResponse(
            content={"title": "", "error": "Conversation has timed out"},
            status_code=401,
        )

    try:
        tree: Tree = await user_manager.get_tree(data.user_id, data.conversation_id)
        title = await tree.create_conversation_title_async()
        return JSONResponse(
            content={"title": title, "error": ""},
            status_code=200,
        )

    except Exception as e:
        logger.exception(f"Error in /title API")
        return JSONResponse(
            content={"title": "", "error": str(e)},
            status_code=200,
        )


# @router.post("/object_relevance")
# async def object_relevance(
#     data: ObjectRelevanceData, user_manager: UserManager = Depends(get_user_manager)
# ):
#     logger.debug(f"/object_relevance API request received")
#     logger.debug(f"User ID: {data.user_id}")
#     logger.debug(f"Conversation ID: {data.conversation_id}")
#     logger.debug(f"Query ID: {data.query_id}")
#     logger.debug(f"Number of objects: {len(data.objects)}")

#     if user_manager.check_tree_timeout(data.user_id, data.conversation_id):
#         logger.warning(
#             f"(/object_relevance) Conversation {data.conversation_id} has timed out for user {data.user_id}"
#         )
#         return JSONResponse(
#             content={"title": "", "error": "Conversation has timed out"},
#             status_code=401,
#         )


#     try:

#         config = user_manager.get_current_user_config(data.user_id)

#         error = ""
#         object_relevance = ObjectRelevanceExecutor()

#         tree = await user_manager.get_tree(data.user_id, data.conversation_id)
#         user_prompt = tree.query_id_to_prompt[data.query_id]
#         prediction = object_relevance(
#             user_prompt=user_prompt,
#             objects=data.objects,
#             lm=load_lm(config.BASE_PROVIDER, config.BASE_MODEL, config.MODEL_API_BASE),
#         )
#         any_relevant = prediction.any_relevant

#     except Exception as e:
#         any_relevant = True
#         error = str(e)

#     return JSONResponse(
#         content={
#             "conversation_id": data.conversation_id,
#             "any_relevant": any_relevant,
#             "error": error,
#         },
#         status_code=200,
#     )


@router.post("/follow_up_suggestions")
async def follow_up_suggestions(
    data: FollowUpSuggestionsData, user_manager: UserManager = Depends(get_user_manager)
):
    logger.debug(f"/follow_up_suggestions API request received")
    logger.debug(f"User ID: {data.user_id}")
    logger.debug(f"Conversation ID: {data.conversation_id}")

    if user_manager.check_tree_timeout(data.user_id, data.conversation_id):
        logger.warning(
            f"(/follow_up_suggestions) Conversation {data.conversation_id} has timed out for user {data.user_id}"
        )
        return JSONResponse(
            content={"suggestions": [], "error": "Conversation has timed out"},
            status_code=408,
        )

    try:
        # wait for tree event to be completed
        local_user = await user_manager.get_user_local(data.user_id)
        event = local_user["tree_manager"].get_event(data.conversation_id)
        # await event.wait()

        # get tree from user_id, conversation_id
        tree: Tree = await user_manager.get_tree(data.user_id, data.conversation_id)

        suggestions = await tree.get_follow_up_suggestions_async()

        event.set()

        return JSONResponse(
            content={"suggestions": suggestions, "error": ""},
            status_code=200,
        )

    except Exception as e:
        logger.exception(f"Error in /follow_up_suggestions API")
        return JSONResponse(
            content={"suggestions": [], "error": str(e)}, status_code=200
        )


@router.post("/debug")
async def debug(data: DebugData, user_manager: UserManager = Depends(get_user_manager)):
    logger.debug(f"/debug API request received")
    logger.debug(f"User ID: {data.user_id}")
    logger.debug(f"Conversation ID: {data.conversation_id}")

    try:
        tree = await user_manager.get_tree(data.user_id, data.conversation_id)
        if tree.debug:
            base_lm = tree.base_lm
            complex_lm = tree.complex_lm

            histories = [None] * 2
            for i, lm in enumerate([base_lm, complex_lm]):
                histories[i] = []
                for lm_history in lm.history:
                    message_thread = []
                    for message in lm_history["messages"]:
                        if isinstance(message["content"], list):
                            # assume its a system message list with one dict element
                            message["content"] = message["content"][0]["text"]
                        message_thread.append(message)

                    message_thread.append(
                        {
                            "role": "assistant",
                            "content": lm_history["response"]
                            .choices[0]
                            .message.content,
                        }
                    )
                    histories[i].append(message_thread)

            out = {
                "base_lm": {"model": base_lm.model, "chat": histories[0]},
                "complex_lm": {"model": complex_lm.model, "chat": histories[1]},
            }
            return JSONResponse(content=out, status_code=200)

        else:
            return JSONResponse(content={}, status_code=200)
    except Exception as e:
        logger.exception(f"Error in /debug API")
        return JSONResponse(
            content={"base_lm": {}, "complex_lm": {}, "error": str(e)}, status_code=200
        )


# @router.post("/get_user_requests")
# async def get_user_requests(
#     data: GetUserRequestsData, user_manager: UserManager = Depends(get_user_manager)
# ):
#     num_requests, max_requests = await user_manager.get_user_requests(data.user_id)
#     return JSONResponse(
#         content={"num_requests": num_requests, "max_requests": max_requests},
#         status_code=200,
#     )



import pytest
from elysia.util.client import ClientManager
from elysia.api.dependencies.common import get_user_manager
from weaviate.util import generate_uuid5
from weaviate.classes.query import Filter
import os
from pathlib import Path
from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv(override=True)


def get_frontend_config_file_paths() -> Path:
    elysia_package_dir = Path(__file__).parent.parent.parent  # Gets to elysia/
    config_dir = elysia_package_dir / "elysia" / "api" / "user_configs"
    config_files = os.listdir(config_dir)
    return [
        f"{config_dir}/{c}"
        for c in config_files
        if c.startswith("frontend_config_test_")
    ]


def pytest_collection_modifyitems(config, items):
    """Skip tests if required environment variables are missing."""
    if (
        ("WCD_URL" not in os.environ and "WEAVIATE_URL" not in os.environ)
        or ("WCD_API_KEY" not in os.environ and "WEAVIATE_API_KEY" not in os.environ)
        or "OPENAI_API_KEY" not in os.environ
        or "OPENROUTER_API_KEY" not in os.environ
    ):
        skip_marker = pytest.mark.skip(
            reason="Missing required database environment variables"
        )
        for item in items:
            item.add_marker(skip_marker)


@pytest.fixture(scope="session", autouse=True)
def cleanup_configs(request):
    yield

    client_manager = ClientManager()

    if not client_manager.is_client:
        return

    # check for local frontend configs
    config_files = get_frontend_config_file_paths()
    for config_file in config_files:
        if os.path.exists(config_file):
            os.remove(config_file)

    # check for weaviate configs
    with client_manager.connect_to_client() as client:
        if client.collections.exists("ELYSIA_CONFIG__"):
            collection = client.collections.get("ELYSIA_CONFIG__")
            test_configs_response = collection.query.fetch_objects(
                limit=1000,
                filters=Filter.all_of(
                    [
                        Filter.by_property("user_id").like("test_*"),
                        Filter.by_property("config_id").like("test_*"),
                    ]
                ),
            )
            for config in test_configs_response.objects:
                collection.data.delete_by_id(config.uuid)

    client_manager.client.close()


@pytest.fixture(scope="session", autouse=True)
def cleanup_collections(request):
    yield

    client_manager = ClientManager()

    if not client_manager.is_client:
        return

    # check for weaviate configs
    with client_manager.connect_to_client() as client:
        for collection_name in client.collections.list_all():
            if collection_name.startswith("Test_ELYSIA_"):
                client.collections.delete(collection_name)
            elif collection_name.startswith("ELYSIA_Test_"):
                client.collections.delete(collection_name)

    client_manager.client.close()


@pytest.fixture(scope="session", autouse=True)
def cleanup_feedbacks(request):
    yield

    client_manager = ClientManager()

    if not client_manager.is_client:
        return

    # check for weaviate configs
    with client_manager.connect_to_client() as client:
        if client.collections.exists("ELYSIA_FEEDBACK__"):
            collection = client.collections.get("ELYSIA_FEEDBACK__")
            test_feedback_response = collection.query.fetch_objects(
                limit=1000,
                filters=Filter.by_property("user_id").like("test_*"),
            )
            for feedback in test_feedback_response.objects:
                collection.data.delete_by_id(feedback.uuid)

    client_manager.client.close()



from typing import Type
from copy import copy

import dspy
from dspy.primitives.module import Module
from dspy.signatures.signature import Signature, ensure_signature
from elysia.tree.objects import TreeData, Atlas
from elysia.util.retrieve_feedback import retrieve_feedback
from elysia.util.client import ClientManager


elysia_meta_prompt = """
You are part of an ensemble of agents that are working together to solve a task.
Your task is just one as part of a larger system of agents designed to work together.
You will be given meta information about where you are in the larger system, 
such as what other tasks have been completed, any items or objects that have been collected during the full process.

Your job is not to answer the entire task, but to complete your part of the task.
Therefore the `user_prompt` should not be judged in its entirety, analyse what you can complete based on the prompt.
Complete this task only, remember that other agents will complete other parts of the task.
"""


class ElysiaChainOfThought(Module):
    """
    A custom reasoning DSPy module that reasons step by step in order to predict the output of a task.
    It will automatically include the most relevant inputs:
    - The user's prompt
    - The conversation history
    - The atlas
    - Any errors (from calls of the same tool)

    And you can also include optional inputs (by setting their boolean flags on initialisation to `True`):
    - The environment
    - The collection schemas
    - The tasks completed

    You can also specify `collection_names` to only include certain collections in the collection schemas.

    It will optionally output (by setting the boolean flags on initialisation to `True`):
    - The reasoning (model step by step reasoning)
    - A message update (if `message_update` is `True`), a brief 'update' message to the user.
    - Whether the task is impossible (boolean)

    You can use this module by calling the `.aforward()` method, passing all your *new* inputs as keyword arguments.
    You do not need to include keyword arguments for the other inputs, like the `environment`.

    Example:
    ```python
    my_module = ElysiaChainOfThought(
        signature=...,
        tree_data=...,
        message_update=True,
        environment=True,
        collection_schemas=True,
        tasks_completed=True,
    )
    my_module.aforward(input1=..., input2=..., lm=...)
    ```
    """

    def __init__(
        self,
        signature: Type[Signature],
        tree_data: TreeData,
        reasoning: bool = True,
        impossible: bool = True,
        message_update: bool = True,
        environment: bool = False,
        collection_schemas: bool = False,
        tasks_completed: bool = False,
        collection_names: list[str] = [],
        **config,
    ):
        """
        Args:
            signature (Type[dspy.Signature]): The signature of the module.
            tree_data (TreeData): Required. The tree data from the Elysia decision tree.
                Used to input the current state of the tree into the prompt.
                If you are using this module as part of a tool, the `tree_data` is an input to the tool call.
            reasoning (bool): Whether to include a reasoning input (chain of thought).
            impossible (bool): Whether to include a boolean flag indicating whether the task is impossible.
                This is useful for stopping the tree from continuing and returning to the base of the decision tree.
                For example, the model judges a query impossible to execute, or the user has not provided enough information.
            message_update (bool): Whether to include a message update input.
                If True, the LLM output will include a brief 'update' message to the user.
                This describes the current action the LLM is performing.
                Designed to increase interactivity and provide the user with information before the final output.
            environment (bool): Whether to include an environment input.
                If True, the module will include the currently stored data from previous tasks and actions into the prompt.
                This is useful so that the LLM knows what has already been done, and can avoid repeating actions.
                Or to use information from the environment to perform the new action.
            collection_schemas (bool): Whether to include a collection schema input.
                If True, the module will include the preprocessed collection schemas in the prompt input.
                This is useful so that the LLM knows the structure of the collections, if querying or similar.
                Use this sparingly, as it will use a large amount of tokens.
                You can specify `collection_names` to only include certain collections in this schema.
            tasks_completed (bool): Whether to include a tasks completed input.
                If True, the module will include the list of tasks completed input.
                This is a nicely formatted list of the tasks that have been completed, with the reasoning for each task.
                This is used so that the LLM has a 'stream of consciousness' of what has already been done,
                as well as to stop it from repeating actions.
                Other information is included in the `tasks_completed` field that format outputs from previous tasks.
                This is useful for continuing a decision logic across tasks, or to reinforce key information.
            collection_names (list[str]): A list of collection names to include in the prompt.
                If provided, this will modify the collection schema input to only include the collections in this list.
                This is useful if you only want to include certain collections in the prompt.
                And to reduce token usage.
            **config (Any): The DSPy configuration for the module.
        """

        super().__init__()

        signature = ensure_signature(signature)  # type: ignore

        # Create a shallow copy of the tree_data
        self.tree_data = copy(tree_data)

        # Note which inputs are required
        self.message_update = message_update
        self.environment = environment
        self.collection_schemas = collection_schemas
        self.tasks_completed = tasks_completed
        self.collection_names = collection_names
        self.reasoning = reasoning
        self.impossible = impossible

        # == Inputs ==

        # -- User Prompt --
        user_prompt_desc = (
            "The user's original question/prompt that needs to be answered. "
            "This, possibly combined with the conversation history, will be used to determine your current action."
        )
        user_prompt_prefix = "${user_prompt}"
        user_prompt_field: str = dspy.InputField(
            prefix=user_prompt_prefix, desc=user_prompt_desc
        )

        # -- Conversation History --
        conversation_history_desc = (
            "Previous messages between user and assistant in chronological order: "
            "[{'role': 'user'|'assistant', 'content': str}] "
            "Use this to maintain conversation context and avoid repetition."
        )
        conversation_history_prefix = "${conversation_history}"
        conversation_history_field: list[dict] = dspy.InputField(
            prefix=conversation_history_prefix, desc=conversation_history_desc
        )

        # -- Atlas --
        atlas_desc = (
            "Your guide to how you should proceed as an agent in this task. "
            "This is pre-defined by the user."
        )
        atlas_prefix = "${atlas}"
        atlas_field: Atlas = dspy.InputField(prefix=atlas_prefix, desc=atlas_desc)

        # -- Errors --
        errors_desc = (
            "Any errors that have occurred during the previous attempt at this action. "
            "This is a list of dictionaries, containing details of the error. "
            "Make an attempt at providing different output to avoid this error now. "
            "If this error is repeated, or you judge it to be unsolvable, you can set `impossible` to True"
        )
        errors_prefix = "${previous_errors}"
        errors_field: list[dict] = dspy.InputField(
            prefix=errors_prefix, desc=errors_desc
        )

        # -- Add to Signature --
        extended_signature = signature.prepend(
            name="user_prompt", field=user_prompt_field, type_=str
        )
        extended_signature = extended_signature.append(
            name="conversation_history",
            field=conversation_history_field,
            type_=list[dict],
        )
        extended_signature = extended_signature.append(
            name="atlas", field=atlas_field, type_=Atlas
        )
        extended_signature = extended_signature.append(
            name="previous_errors", field=errors_field, type_=list[dict]
        )

        # == Optional Inputs / Outputs ==

        # -- Environment Input --
        if environment:
            environment_desc = (
                "Information gathered from completed tasks. "
                "Empty if no data has been retrieved yet. "
                "Use to determine if more information is needed. "
                "Additionally, use this as a reference to determine if you have already completed a task/what items are already available, to avoid repeating actions. "
                "All items here are already shown to the user, so do not repeat information from these fields unless summarising, providing extra information or otherwise. "
                "E.g., do not list out anything from here, only provide new content to the user."
            )
            environment_prefix = "${environment}"
            environment_field: dict = dspy.InputField(
                prefix=environment_prefix, desc=environment_desc
            )
            extended_signature = extended_signature.append(
                name="environment", field=environment_field, type_=dict
            )

        # -- Collection Schema Input --
        if collection_schemas:
            collection_schemas_desc = (
                "Metadata about available collections and their schemas: "
                "This is a dictionary with the following fields: "
                "{\n"
                "    name: collection name,\n"
                "    length: number of objects in the collection,\n"
                "    summary: summary of the collection,\n"
                "    fields: [\n"
                "        {\n"
                "            name: field_name,\n"
                "            groups: a dict with the value and count of each group.\n"
                "                a comprehensive list of all unique values that exist in the field.\n"
                "                if this is None, then no relevant groups were found.\n"
                "                these values are string, but the actual values in the collection are the 'type' of the field.\n"
                "            mean: mean of the field. if the field is text, this refers to the means length (in tokens) of the texts in this field. if the type is a list, this refers to the mean length of the lists,\n"
                "            range: minimum and maximum values of the length,\n"
                "            type: the data type of the field.\n"
                "        },\n"
                "        ...\n"
                "    ]\n"
                "}\n"
            )
            collection_schemas_prefix = "${collection_schemas}"
            collection_schemas_field: dict = dspy.InputField(
                prefix=collection_schemas_prefix, desc=collection_schemas_desc
            )
            extended_signature = extended_signature.append(
                name="collection_schemas", field=collection_schemas_field, type_=dict
            )

        # -- Tasks Completed Input --
        if tasks_completed:
            tasks_completed_desc = (
                "Which tasks have been completed in order. "
                "These are numbered so that higher numbers are more recent. "
                "Separated by prompts (so you should identify the prompt you are currently working on to see what tasks have been completed so far) "
                "Also includes reasoning for each task, to continue a decision logic across tasks. "
                "Use this to determine whether future searches, for this prompt are necessary, and what task(s) to choose. "
                "It is IMPORTANT that you separate what actions have been completed for which prompt, so you do not think you have failed an attempt for a different prompt."
            )
            tasks_completed_prefix = "${tasks_completed}"
            tasks_completed_field: str = dspy.InputField(
                prefix=tasks_completed_prefix, desc=tasks_completed_desc
            )
            extended_signature = extended_signature.append(
                name="tasks_completed", field=tasks_completed_field, type_=str
            )

        # -- Impossible Field --
        if impossible:
            impossible_desc = (
                "Given the actions you have available, and the environment/information. "
                "Is the task impossible to complete? "
                "I.e., do you wish that you had a different task to perform/choose from and hence should return to the base of the decision tree?"
                "Do not base this judgement on the entire prompt, as it is possible that other agents can perform other aspects of the request."
                "Do not judge impossibility based on if tasks have been completed, only on the current action and environment."
            )
            impossible_prefix = "${impossible}"
            impossible_field: bool = dspy.OutputField(
                prefix=impossible_prefix, desc=impossible_desc
            )
            extended_signature = extended_signature.prepend(
                name="impossible", field=impossible_field, type_=bool
            )

        # -- Message Update Output --
        if message_update:
            message_update_desc = (
                "Continue your current message to the user "
                "(latest assistant field in conversation history) with ONE concise sentence that: "
                "- Describes NEW technical details about your latest action "
                "- Highlights specific parameters or logic you just applied "
                "- Avoids repeating anything from conversation history "
                "- Speaks directly to them (no 'the user'), gender neutral message "
                "Just provide the new sentence update, not the full message from the conversation history. "
                "Your response should be based on only the part of the user's request that you can work on. "
                "It is possible other agents can perform other aspects of the request, so do not respond as if you cannot complete the entire request."
            )

            message_update_prefix = "${message_update}"
            message_update_field: str = dspy.OutputField(
                prefix=message_update_prefix, desc=message_update_desc
            )
            extended_signature = extended_signature.prepend(
                name="message_update", field=message_update_field, type_=str
            )

        # -- Reasoning Field --
        if reasoning:
            reasoning_desc = (
                "Reasoning: Repeat relevant parts of the any context within your environment, "
                "Evaluate all relevant information from the inputs, including any previous errors if applicable, "
                "use this to think step by step in order to answer the query."
                "Limit your reasoning to maximum 150 words. Only exceed this if the task is very complex."
            )
            reasoning_prefix = "${reasoning}"
            reasoning_field: str = dspy.OutputField(
                prefix=reasoning_prefix, desc=reasoning_desc
            )
            extended_signature = extended_signature.prepend(
                name="reasoning", field=reasoning_field, type_=str
            )

        # -- Predict --
        self.predict = dspy.Predict(extended_signature, **config)
        self.predict.signature.instructions += elysia_meta_prompt  # type: ignore

    def _add_tree_data_inputs(self, kwargs: dict):

        # Add the tree data inputs to the kwargs
        kwargs["user_prompt"] = self.tree_data.user_prompt
        kwargs["conversation_history"] = self.tree_data.conversation_history
        kwargs["atlas"] = self.tree_data.atlas
        kwargs["previous_errors"] = self.tree_data.get_errors()

        # Add the optional inputs to the kwargs
        if self.environment:
            kwargs["environment"] = self.tree_data.environment.environment

        if self.collection_schemas:
            if self.collection_names != []:
                kwargs["collection_schemas"] = (
                    self.tree_data.output_collection_metadata(
                        collection_names=self.collection_names, with_mappings=False
                    )
                )
            else:
                kwargs["collection_schemas"] = (
                    self.tree_data.output_collection_metadata(with_mappings=False)
                )

        if self.tasks_completed:
            kwargs["tasks_completed"] = self.tree_data.tasks_completed_string()

        return kwargs

    def forward(self, **kwargs):
        kwargs = self._add_tree_data_inputs(kwargs)
        return self.predict(**kwargs)

    async def aforward(self, **kwargs):
        kwargs = self._add_tree_data_inputs(kwargs)
        return await self.predict.acall(**kwargs)

    async def aforward_with_feedback_examples(
        self,
        feedback_model: str,
        client_manager: ClientManager,
        base_lm: dspy.LM,
        complex_lm: dspy.LM,
        num_base_lm_examples: int = 3,
        return_example_uuids: bool = False,
        **kwargs,
    ) -> tuple[dspy.Prediction, list[str]] | dspy.Prediction:
        """
        Use the forward pass of the module with feedback examples.
        This will first retrieve examples from the feedback collection, and use those as few-shot examples to run the module.
        It retrieves based from vectorising and searching on the user's prompt, finding similar prompts from the feedback collection.
        This is an EXPERIMENTAL feature, and may not work as expected.

        If the number of examples is less than `num_base_lm_examples`, the module will use the complex LM.
        Otherwise, it will use the base LM. This is so that the less accurate, but faster base LM can be used when guidance is available.
        However, when there are insufficient examples, the complex LM will be used.

        Args:
            feedback_model (str): The label of the feedback data to use as examples.
                E.g., "decision" is the default name given to examples for the LM in the decision tree.
                This is used to retrieve the examples from the feedback collection.
            client_manager (ClientManager): The client manager to use.
            base_lm (dspy.LM): The base LM to (conditionally) use.
            complex_lm (dspy.LM): The complex LM to (conditionally) use.
            num_base_lm_examples (int): The threshold number of examples to use the base LM.
                When there are fewer examples than this, the complex LM will be used.
            **kwargs (Any): The keyword arguments to pass to the forward pass.
                Important: All additional inputs to the DSPy module should be passed here as keyword arguments.
                Also: Do not include `lm` in the kwargs, as this will be set automatically.

        Returns:
            (dspy.Prediction): The prediction from the forward pass.
        """

        examples, uuids = await retrieve_feedback(
            client_manager, self.tree_data.user_prompt, feedback_model, n=10
        )
        if len(examples) > 0:
            optimizer = dspy.LabeledFewShot(k=10)
            optimized_module = optimizer.compile(self, trainset=examples)
        else:
            if return_example_uuids:
                return (
                    await self.aforward(lm=complex_lm, **kwargs),
                    uuids,
                )
            else:
                return await self.aforward(lm=complex_lm, **kwargs)

        # Select the LM to use based on the number of examples
        if len(examples) < num_base_lm_examples:
            if return_example_uuids:
                return (
                    await optimized_module.aforward(lm=complex_lm, **kwargs),
                    uuids,
                )
            else:
                return await optimized_module.aforward(lm=complex_lm, **kwargs)
        else:
            if return_example_uuids:
                return (
                    await optimized_module.aforward(lm=base_lm, **kwargs),
                    uuids,
                )
            else:
                return await optimized_module.aforward(lm=base_lm, **kwargs)



import pytest
import random
from fastapi.responses import JSONResponse

import asyncio
import json

from elysia.api.api_types import (
    ViewPaginatedCollectionData,
    UpdateCollectionMetadataData,
    MetadataNamedVectorData,
    MetadataFieldData,
    InitialiseTreeData,
)
from elysia.api.dependencies.common import get_user_manager
from elysia.api.routes.collections import (
    collections_list,
    view_paginated_collection,
    collection_metadata,
    get_object,
    mapping_types,
    update_metadata,
    delete_metadata,
)
from elysia.api.routes.init import initialise_user, initialise_tree
from elysia.api.core.log import logger, set_log_level
from elysia.preprocessing.collection import preprocess_async
from elysia.util.client import ClientManager

set_log_level("CRITICAL")

from weaviate.classes.config import Configure
import weaviate.classes as wvc
from weaviate.util import generate_uuid5


async def initialise_user_and_tree(user_id: str, conversation_id: str):
    user_manager = get_user_manager()

    response = await initialise_user(
        user_id,
        user_manager,
    )

    response = await initialise_tree(
        user_id,
        conversation_id,
        InitialiseTreeData(
            low_memory=False,
        ),
        user_manager,
    )


def create_named_vectors_collection(
    client_manager: ClientManager, collection_name: str
):

    with client_manager.connect_to_client() as client:

        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)

        collection = client.collections.create(
            name=collection_name,
            properties=[
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(
                    name="category", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(name="author", data_type=wvc.config.DataType.TEXT),
            ],
            vector_config=[
                wvc.config.Configure.Vectors.text2vec_openai(
                    name="content_vector", source_properties=["content"]
                ),
                wvc.config.Configure.Vectors.text2vec_openai(
                    name="title_vector", source_properties=["title"]
                ),
                wvc.config.Configure.Vectors.self_provided(name="custom_vector"),
            ],
        )

        documents_data = [
            {
                "product_id": "doc1",
                "title": "Introduction to Machine Learning",
                "content": "Machine learning is a subset of artificial intelligence that focuses on algorithms that can learn from data.",
                "category": "Education",
                "author": "Dr. Smith",
                "custom_vector": [0.1, 0.2, 0.3, 0.4, 0.5],  # Custom embedding
            },
            {
                "product_id": "doc2",
                "title": "Python Programming Basics",
                "content": "Python is a versatile programming language used in web development, data science, and automation.",
                "category": "Programming",
                "author": "Jane Doe",
                "custom_vector": [0.6, 0.7, 0.8, 0.9, 1.0],
            },
            {
                "product_id": "doc3",
                "title": "Database Design Principles",
                "content": "Good database design involves normalization, indexing, and understanding relationships between entities.",
                "category": "Database",
                "author": "Bob Johnson",
                "custom_vector": [0.2, 0.4, 0.6, 0.8, 0.3],
            },
        ]

        # Import data with named vectors
        with collection.batch.fixed_size(batch_size=10) as batch:
            for doc in documents_data:
                batch.add_object(
                    properties={
                        "title": doc["title"],
                        "content": doc["content"],
                        "category": doc["category"],
                        "author": doc["author"],
                    },
                    vector={"custom_vector": doc["custom_vector"]},
                    uuid=generate_uuid5(doc["product_id"]),
                )


def create_regular_vectorizer_collection(
    client_manager: ClientManager, collection_name: str
):

    with client_manager.connect_to_client() as client:

        if client.collections.exists(collection_name):
            client.collections.delete(collection_name)

        collection = client.collections.create(
            name=collection_name,
            properties=[
                wvc.config.Property(
                    name="product_id", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(
                    name="description", data_type=wvc.config.DataType.TEXT
                ),
                wvc.config.Property(
                    name="tags", data_type=wvc.config.DataType.TEXT_ARRAY
                ),
                wvc.config.Property(name="price", data_type=wvc.config.DataType.NUMBER),
            ],
            vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_openai(
                vectorize_collection_name=False
            ),
        )

        products_data = [
            {
                "product_id": "prod1",
                "title": "Wireless Bluetooth Headphones",
                "description": "High-quality wireless headphones with noise cancellation and 30-hour battery life.",
                "tags": ["electronics", "audio", "wireless", "bluetooth"],
                "price": 199.99,
            },
            {
                "product_id": "prod2",
                "title": "Organic Green Tea",
                "description": "Premium organic green tea sourced from high-altitude tea gardens in Japan.",
                "tags": ["beverage", "tea", "organic", "healthy"],
                "price": 24.99,
            },
            {
                "product_id": "prod3",
                "title": "Yoga Mat with Alignment Lines",
                "description": "Non-slip yoga mat with helpful alignment lines for proper posture during practice.",
                "tags": ["fitness", "yoga", "exercise", "wellness"],
                "price": 39.99,
            },
            {
                "product_id": "prod4",
                "title": "Smart Home Security Camera",
                "description": "WiFi-enabled security camera with motion detection and smartphone notifications.",
                "tags": ["security", "smart-home", "camera", "surveillance"],
                "price": 149.99,
            },
        ]

        with collection.batch.fixed_size(batch_size=10) as batch:
            for product in products_data:
                batch.add_object(
                    properties={
                        "title": product["title"],
                        "description": product["description"],
                        "tags": product["tags"],
                        "price": product["price"],
                    },
                    uuid=generate_uuid5(product["product_id"]),
                )


def read_response(response: JSONResponse):
    return json.loads(response.body)


@pytest.mark.asyncio
async def test_collections():

    user_id = "test_user_collections"
    conversation_id = "test_conversation_collections"
    collection_name = "Test_ELYSIA_collection_collections"

    client_manager = ClientManager()
    create_regular_vectorizer_collection(client_manager, collection_name)

    user_manager = get_user_manager()
    await initialise_user_and_tree(user_id, conversation_id)
    basic = await collections_list(user_id, user_manager)
    basic = read_response(basic)
    assert basic["error"] == "", basic["error"]
    assert collection_name in [
        collection["name"] for collection in basic["collections"]
    ]


@pytest.mark.asyncio
async def test_get_object():

    user_id = "test_user_get_object"
    conversation_id = "test_conversation_get_object"
    collection_name = "Test_ELYSIA_collection_get_object"

    client_manager = ClientManager()
    create_regular_vectorizer_collection(client_manager, collection_name)

    user_manager = get_user_manager()
    await initialise_user_and_tree(user_id, conversation_id)

    # first manually get a UUID
    user_local = await user_manager.get_user_local(user_id)
    client_manager = user_local["client_manager"]
    async with client_manager.connect_to_async_client() as client:
        collection = client.collections.get(collection_name)
        response = await collection.query.fetch_objects(limit=1)
        uuid = str(response.objects[0].uuid)

    # test with valid UUID
    basic = await get_object(
        user_id,
        collection_name=collection_name,
        uuid=uuid,
        user_manager=user_manager,
    )

    basic = read_response(basic)
    assert basic["error"] == ""

    # test with invalid UUID
    invalid = await get_object(
        user_id,
        collection_name=collection_name,
        uuid="invalid",
        user_manager=user_manager,
    )

    invalid = read_response(invalid)
    assert invalid["error"] != ""


@pytest.mark.asyncio
async def test_view_paginated_collection():

    user_id = "test_user_view_paginated_collection"
    conversation_id = "test_conversation_view_paginated_collection"
    collection_name = "Test_ELYSIA_collection_view_paginated_collection"
    client_manager = ClientManager()
    create_regular_vectorizer_collection(client_manager, collection_name)

    user_manager = get_user_manager()
    await initialise_user_and_tree(user_id, conversation_id)

    basic = await view_paginated_collection(
        user_id,
        collection_name=collection_name,
        data=ViewPaginatedCollectionData(
            query="",
            page_size=50,
            page_number=1,
            sort_on="title",
            ascending=True,
            filter_config={},
        ),
        user_manager=user_manager,
    )
    basic = read_response(basic)
    assert basic["error"] == ""

    # test with filter
    filter = await view_paginated_collection(
        user_id,
        collection_name=collection_name,
        data=ViewPaginatedCollectionData(
            page_size=50,
            page_number=1,
            sort_on="title",
            ascending=True,
            filter_config={
                "type": "all",
                "filters": [
                    {
                        "field": "product_id",
                        "operator": "equal",
                        "value": "prod2",
                    }
                ],
            },
        ),
        user_manager=user_manager,
    )
    filter = read_response(filter)
    assert filter["error"] == ""

    # test with query
    query = await view_paginated_collection(
        user_id,
        collection_name=collection_name,
        data=ViewPaginatedCollectionData(
            query="hello",
            page_size=50,
            page_number=1,
            ascending=True,
            filter_config={},
        ),
        user_manager=user_manager,
    )
    query = read_response(query)
    assert query["error"] == ""

    # test with out of bounds page number, should return empty list
    out_of_bounds = await view_paginated_collection(
        user_id,
        collection_name=collection_name,
        data=ViewPaginatedCollectionData(
            page_size=50,
            page_number=100,
            sort_on="title",
            ascending=True,
            filter_config={},
        ),
        user_manager=user_manager,
    )
    out_of_bounds = read_response(out_of_bounds)
    assert out_of_bounds["error"] == ""
    assert len(out_of_bounds["items"]) == 0


@pytest.mark.asyncio
async def test_mapping_types():
    await mapping_types()



import asyncio
import datetime
import os
import threading

from contextlib import asynccontextmanager, contextmanager
from typing import AsyncGenerator, Generator, Any
from logging import Logger

import weaviate
from weaviate.classes.init import Auth
from weaviate.client import WeaviateClient, WeaviateAsyncClient
from elysia.config import settings as environment_settings, Settings

api_key_map = {
    # Regular API keys
    "ANTHROPIC_APIKEY": "X-Anthropic-Api-Key",
    "ANYSCALE_APIKEY": "X-Anyscale-Api-Key",
    "AWS_ACCESS_KEY": "X-Aws-Access-Key",
    "AWS_SECRET_KEY": "X-Aws-Secret-Key",
    "COHERE_API_KEY": "X-Cohere-Api-Key",
    "DATABRICKS_TOKEN": "X-Databricks-Token",
    "FRIENDLI_TOKEN": "X-Friendli-Api-Key",
    "VERTEX_APIKEY": "X-Goog-Vertex-Api-Key",
    "STUDIO_APIKEY": "X-Goog-Studio-Api-Key",
    "HUGGINGFACE_APIKEY": "X-HuggingFace-Api-Key",
    "JINAAI_APIKEY": "X-JinaAI-Api-Key",
    "MISTRAL_APIKEY": "X-Mistral-Api-Key",
    "NVIDIA_APIKEY": "X-Nvidia-Api-Key",
    "OPENAI_APIKEY": "X-OpenAI-Api-Key",
    "AZURE_APIKEY": "X-Azure-Api-Key",
    "VOYAGE_APIKEY": "X-Voyage-Api-Key",
    "XAI_APIKEY": "X-Xai-Api-Key",
    # And separate out "API_KEY"
    "ANTHROPIC_API_KEY": "X-Anthropic-Api-Key",
    "ANYSCALE_API_KEY": "X-Anyscale-Api-Key",
    "AWS_ACCESS_KEY": "X-Aws-Access-Key",
    "AWS_SECRET_KEY": "X-Aws-Secret-Key",
    "COHERE_API_KEY": "X-Cohere-Api-Key",
    "DATABRICKS_TOKEN": "X-Databricks-Token",
    "FRIENDLI_TOKEN": "X-Friendli-Api-Key",
    "VERTEX_API_KEY": "X-Goog-Vertex-Api-Key",
    "STUDIO_API_KEY": "X-Goog-Studio-Api-Key",
    "HUGGINGFACE_API_KEY": "X-HuggingFace-Api-Key",
    "JINAAI_API_KEY": "X-JinaAI-Api-Key",
    "MISTRAL_API_KEY": "X-Mistral-Api-Key",
    "NVIDIA_API_KEY": "X-Nvidia-Api-Key",
    "OPENAI_API_KEY": "X-OpenAI-Api-Key",
    "AZURE_API_KEY": "X-Azure-Api-Key",
    "VOYAGE_API_KEY": "X-Voyage-Api-Key",
    "XAI_API_KEY": "X-Xai-Api-Key",
}


class ClientManager:
    """
    Handles the creation and management of the Weaviate client.
    Handles cases where the client can be used in more than one thread or async operation at a time,
    via threading and asyncio locks.
    Also can use methods for restarting client if its been inactive.
    """

    def __init__(
        self,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
        client_timeout: datetime.timedelta | int | None = None,
        logger: Logger | None = None,
        settings: Settings | None = None,
        **kwargs,
    ) -> None:
        """
        Args:
            wcd_url (str): the url of the Weaviate cluster. Defaults to global settings config.
            wcd_api_key (str): the api key for the Weaviate cluster. Defaults to global settings config.
            client_timeout (datetime.timedelta | int | None): how long (in minutes) means the client should be restarted. Defaults to 3 minutes.
            logger (Logger | None): a logger object for logging messages. Defaults to None.
            settings (Settings | None): a settings object for the client manager. Defaults to environment settings.
            **kwargs (Any): any other api keys for third party services (formatted as e.g. OPENAI_APIKEY).

        Example:
        ```python
        client_manager = ClientManager(
            wcd_url="https://my-weaviate-cluster...",
            wcd_api_key="my-api-key...",
            OPENAI_APIKEY="my-openai-api-key...",
            HUGGINGFACE_APIKEY="my-huggingface-api-key...",
        )
        ```
        """

        self.logger = logger

        if client_timeout is None:
            self.client_timeout = datetime.timedelta(
                minutes=int(os.getenv("CLIENT_TIMEOUT", 3))
            )
        elif isinstance(client_timeout, int):
            self.client_timeout = datetime.timedelta(minutes=client_timeout)
        else:
            self.client_timeout = client_timeout

        if settings is None:
            self.settings = environment_settings
        else:
            self.settings = settings

        # Set the weaviate cluster url and api key
        if wcd_url is None:
            self.wcd_url = self.settings.WCD_URL
        else:
            self.wcd_url = wcd_url

        if wcd_api_key is None:
            self.wcd_api_key = self.settings.WCD_API_KEY
        else:
            self.wcd_api_key = wcd_api_key

        # Set the api keys for non weaviate cluster (third parties)
        self.headers = {}
        for api_key in self.settings.API_KEYS:
            if api_key.lower() in [a.lower() for a in api_key_map.keys()]:
                self.headers[api_key_map[api_key.upper()]] = self.settings.API_KEYS[
                    api_key
                ]

        # From kwargs
        for kwarg in kwargs:
            if kwarg.lower() in [a.lower() for a in api_key_map.keys()]:
                self.headers[api_key_map[kwarg.upper()]] = kwargs[kwarg]

        # Create locks for client events
        self.async_lock = asyncio.Lock()
        self.sync_lock = threading.Lock()

        # In use counter tracks when the client is in use and by how many operations. 0 = can restart
        self.async_in_use_counter = 0
        self.sync_in_use_counter = 0
        self.async_restart_event = asyncio.Event()
        self.sync_restart_event = threading.Event()

        self.last_used_sync_client = datetime.datetime.now()
        self.last_used_async_client = datetime.datetime.now()

        self.async_client = None
        self.async_init_completed = False
        self.is_client = self.wcd_url != "" and self.wcd_api_key != ""

        if self.logger:
            if self.wcd_api_key == "" and self.wcd_url == "":
                self.logger.warning(
                    "WCD_URL and WCD_API_KEY are not set. "
                    "All Weaviate functionality will be disabled."
                )
            elif self.wcd_url == "":
                self.logger.warning(
                    "WCD_URL is not set. "
                    "All Weaviate functionality will be disabled."
                )
            elif self.wcd_api_key == "":
                self.logger.warning(
                    "WCD_API_KEY is not set. "
                    "All Weaviate functionality will be disabled."
                )
            else:
                self.logger.debug(
                    "Weaviate client initialised. "
                    "All Weaviate functionality will be enabled."
                )

        if not self.is_client:
            return

        # Start sync client
        self.client = self.get_client()
        self.sync_restart_event.set()

    async def reset_keys(
        self,
        wcd_url: str | None = None,
        wcd_api_key: str | None = None,
        api_keys: dict[str, str] = {},
    ) -> None:
        """
        Set the API keys, WCD_URL and WCD_API_KEY from the settings object.

        Args:
            wcd_url (str): the url of the Weaviate cluster.
            wcd_api_key (str): the api key for the Weaviate cluster.
            api_keys (dict): a dictionary of api keys for third party services.
        """
        self.wcd_url = wcd_url
        self.wcd_api_key = wcd_api_key

        self.headers = {}

        for api_key in api_keys:
            if api_key.lower() in [a.lower() for a in api_key_map.keys()]:
                self.headers[api_key_map[api_key.upper()]] = api_keys[api_key]

        self.is_client = self.wcd_url != "" and self.wcd_api_key != ""
        if self.is_client:
            await self.restart_client(force=True)
            await self.restart_async_client(force=True)
            await self.start_clients()

    async def start_clients(self) -> None:
        """
        Start the async and sync clients if they are not already running.
        """

        if not self.is_client:
            raise ValueError(
                "Weaviate is not available. Please set the WCD_URL and WCD_API_KEY in the settings."
            )

        if self.async_client is None:
            self.async_client = await self.get_async_client()
            self.async_restart_event.set()

        if not self.async_client.is_connected():
            await self.async_client.connect()

        self.async_init_completed = True

        if not self.client.is_connected():
            self.client.connect()

    def update_last_user_request(self) -> None:
        self.last_user_request = datetime.datetime.now()

    def update_last_used_sync_client(self) -> None:
        self.last_used_sync_client = datetime.datetime.now()

    def update_last_used_async_client(self) -> None:
        self.last_used_async_client = datetime.datetime.now()

    def get_client(self) -> WeaviateClient:
        if self.wcd_url is None or self.wcd_api_key is None:
            raise ValueError("WCD_URL and WCD_API_KEY must be set")

        return weaviate.connect_to_weaviate_cloud(
            cluster_url=self.wcd_url,
            auth_credentials=Auth.api_key(self.wcd_api_key),
            headers=self.headers,
            skip_init_checks=True,
        )

    async def get_async_client(self) -> WeaviateAsyncClient:
        if self.wcd_url is None or self.wcd_api_key is None:
            raise ValueError("WCD_URL and WCD_API_KEY must be set")

        return weaviate.use_async_with_weaviate_cloud(
            cluster_url=self.wcd_url,
            auth_credentials=Auth.api_key(self.wcd_api_key),
            headers=self.headers,
            skip_init_checks=True,
        )

    @contextmanager
    def connect_to_client(self) -> Generator[WeaviateClient, Any, None]:
        """
        A context manager to connect to the _sync_ client.

        E.g.

        ```python
        with client_manager.connect_to_client():
            # do stuff with the weaviate client
            ...
        ```
        """
        if not self.is_client:
            raise ValueError(
                "Weaviate is not available. Please set the WCD_URL and WCD_API_KEY in the settings."
            )

        self.sync_restart_event.wait()
        with self.sync_lock:
            self.sync_in_use_counter += 1

        if not self.client.is_connected():
            self.client.connect()

        connection = _ClientConnection(self, self.client)
        with connection:
            yield connection.client

    @asynccontextmanager
    async def connect_to_async_client(
        self,
    ) -> AsyncGenerator[WeaviateAsyncClient, Any]:
        """
        A context manager to connect to the _async_ client.

        E.g.
        ```python
        async with client_manager.connect_to_async_client():
            # do stuff with the async weaviate client
            ...
        ```
        """
        if not self.is_client:
            raise ValueError(
                "Weaviate is not available. Please set the WCD_URL and WCD_API_KEY in the settings."
            )

        if not self.async_init_completed:
            await self.start_clients()

        await self.async_restart_event.wait()
        async with self.async_lock:
            self.async_in_use_counter += 1

        if self.async_client is None:
            raise ValueError("Async client not initialised")

        if not self.async_client.is_connected():
            await self.async_client.connect()

        connection = _AsyncClientConnection(self, self.async_client)
        async with connection:
            yield connection.client

    async def restart_async_client(self, force=False) -> None:
        """
        Restart the async client if it has not been used in the last client_timeout minutes (set in init).
        """
        if self.client_timeout == datetime.timedelta(minutes=0) and not force:
            return

        # First check if the client has been used in the last X minutes
        if (
            datetime.datetime.now() - self.last_used_async_client > self.client_timeout
            or force
        ):
            # Acquire lock before modifying shared state to prevent race conditions
            try:
                async with self.async_lock:
                    # Clear the event WHILE holding the lock to prevent new connections from starting
                    # This ensures no new connections can proceed until we're done
                    self.async_restart_event.clear()

                    # Record current counter value before waiting
                    last_recorded_counter = self.async_in_use_counter

                    # Set reasonable timeout values
                    time_spent = 0
                    max_wait_time = 10  # seconds
                    check_interval = 0.1  # seconds

                    # Only wait if there are active connections
                    if last_recorded_counter > 0:
                        # Wait for existing connections to complete
                        while self.async_in_use_counter > 0:
                            # Release lock during sleep to prevent deadlock
                            self.async_lock.release()

                            # Only timeout after 10 seconds if nothing is happening
                            # if the counter is changing, then things are happening and we should wait
                            if self.async_in_use_counter != last_recorded_counter:
                                last_recorded_counter = self.async_in_use_counter
                                time_spent = 0

                            try:
                                await asyncio.sleep(check_interval)
                                time_spent += check_interval
                                if time_spent > max_wait_time:
                                    if self.logger:
                                        self.logger.error(
                                            f"Async client restart timed out after {max_wait_time} seconds. "
                                        )
                                    break
                            finally:
                                # Re-acquire lock after sleep
                                await self.async_lock.acquire()

                    # Handle timeout case - must reset state regardless of timeout
                    if self.async_in_use_counter > 0:
                        if self.logger:
                            self.logger.error(
                                "Force resetting async client state due to timeout"
                            )
                        self.async_in_use_counter = 0

                    # Whether we timed out or not, we need to restart the client
                    try:
                        # Only close if client exists and is connected
                        if (
                            hasattr(self, "async_client")
                            and self.async_client is not None
                        ):
                            await self.async_client.close()
                        await asyncio.sleep(0.1)
                        # Create a new client instance
                        self.async_client = await self.get_async_client()
                    except Exception as e:
                        if self.logger:
                            self.logger.error(
                                f"Error during async client restart: {str(e)}"
                            )
                        # Create a new client anyway to ensure we have a valid client
                        self.async_client = await self.get_async_client()
                    finally:
                        # CRITICAL: Always set the event to prevent deadlocks
                        # This ensures waiting connections can proceed
                        self.async_restart_event.set()

            except Exception as e:
                if self.logger:
                    self.logger.error(
                        f"Unexpected error in async client restart: {str(e)}"
                    )
                # Ensure the event is set in all error cases
                self.async_restart_event.set()
                # Attempt to create a new client
                self.async_client = await self.get_async_client()

    async def restart_client(self, force=False) -> None:
        """
        Restart the sync client if it has not been used in the last client_timeout minutes (set in init).
        """
        if self.client_timeout == datetime.timedelta(minutes=0) and not force:
            return

        # First check if the client has been used in the last X minutes
        if (
            datetime.datetime.now() - self.last_used_sync_client > self.client_timeout
            or force
        ):
            # Use both locks to prevent any race conditions between sync and async operations
            try:
                # Acquire sync lock first
                with self.sync_lock:
                    # Clear event while holding the lock to prevent race conditions
                    self.sync_restart_event.clear()

                    # Record current counter value
                    last_recorded_counter = self.sync_in_use_counter

                    # Set reasonable timeout values
                    time_spent = 0
                    max_wait_time = 10  # seconds
                    check_interval = 0.1  # seconds

                    # Only wait if there are active connections
                    if last_recorded_counter > 0:
                        # Wait for existing connections to complete
                        while self.sync_in_use_counter > 0:
                            # Must release lock during async sleep to prevent deadlock
                            self.sync_lock.release()

                            # Only timeout after 10 seconds if nothing is happening
                            # if the counter is changing, then things are happening and we should wait
                            if self.sync_in_use_counter != last_recorded_counter:
                                last_recorded_counter = self.sync_in_use_counter
                                time_spent = 0

                            try:
                                await asyncio.sleep(check_interval)
                                time_spent += check_interval
                                if time_spent > max_wait_time:
                                    if self.logger:
                                        self.logger.error(
                                            f"Sync client restart timed out after {max_wait_time}s. "
                                            f"Initial counter: {last_recorded_counter}, Current: {self.sync_in_use_counter}"
                                        )
                                    break
                            finally:
                                # Re-acquire lock
                                self.sync_lock.acquire()

                    # Handle timeout case - must reset state regardless of timeout
                    if self.sync_in_use_counter > 0:
                        if self.logger:
                            self.logger.error(
                                "Force resetting sync client state due to timeout"
                            )
                        self.sync_in_use_counter = 0

                    # Whether we timed out or not, we need to restart the client
                    try:
                        # Only close if client exists and is connected
                        if hasattr(self, "client") and self.client is not None:
                            self.client.close()
                        await asyncio.sleep(0.1)
                        # Create a new client instance
                        self.client = self.get_client()
                    except Exception as e:
                        if self.logger:
                            self.logger.error(
                                f"Error during sync client restart: {str(e)}"
                            )
                        # Create a new client anyway
                        self.client = self.get_client()
                    finally:
                        # CRITICAL: Always set the event to prevent deadlocks
                        self.sync_restart_event.set()

            except Exception as e:
                if self.logger:
                    self.logger.error(
                        f"Unexpected error in sync client restart: {str(e)}"
                    )
                # Ensure the event is set in all error cases
                self.sync_restart_event.set()
                # Attempt to create a new client
                self.client = self.get_client()

    async def close_clients(self) -> None:
        """
        Close both the async and sync clients.
        Should not be called inside a Tool or other function inside the decision tree.
        """
        if hasattr(self, "async_client") and self.async_client is not None:
            await self.async_client.close()
        if hasattr(self, "client") and self.client is not None:
            self.client.close()


# Custom context managers so that clients do not close after use (instead on a timer)
class _ClientConnection:
    def __init__(self, manager: ClientManager, client: WeaviateClient):
        self.manager = manager
        self.client = client

    def __enter__(self) -> WeaviateClient:
        return self.client

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        with self.manager.sync_lock:
            self.manager.sync_in_use_counter -= 1
        self.manager.update_last_used_sync_client()


class _AsyncClientConnection:
    def __init__(self, manager: ClientManager, client: WeaviateAsyncClient):
        self.manager = manager
        self.client = client

    async def __aenter__(self) -> WeaviateAsyncClient:
        return self.client

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        async with self.manager.async_lock:
            self.manager.async_in_use_counter -= 1
        self.manager.update_last_used_async_client()
